#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu

# kill subprocesses when the caller script terminates
trap "trap - SIGTERM && jobs -p | xargs -r kill &>/dev/null" INT TERM EXIT


#GLOBALS:
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error_with_code "10001"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error_with_code "10002"
fi

readonly SCRIPT_DIR
readonly MODULES_DIR

# Set logfile path if not already set by caller
: "${INSTALLER_LOGFILE:="${SCRIPT_DIR}/install-$(date +'%Y-%m-%dT%H_%M_%S').log"}"

# Redirect /dev/null to INSTALLER_LOGFILE if set
dev_null="/dev/null"
[[ -n "${INSTALLER_LOGFILE:-}" ]] && dev_null="${INSTALLER_LOGFILE}"

## Locale
LANG_INFRA_STARTED_PREFIX="[Infrastructure] [Started]"
LANG_INFRA_FAILED_PREFIX="[Infrastructure] [Failed]"
LANG_INFRA_COMPLETED_PREFIX="[Infrastructure] [Completed]"
# LANG_INFRA_SKIPPED_PREFIX="[Infrastructure] [Skipped]"
LANG_K8S_INSTALL_MESSAGE="Installing and configuring kubernetes"
LANG_GPU_SETUP_MESSAGE="Configuring the GPU"
LANG_STORAGE_SETUP_MESSAGE="Installing and configuring storage"
LANG_ROUTING_SETUP_MESSAGE="Installing and configuring routing"
LANG_COREDNS_SETUP_MESSAGE="Configuring custom RKE2 CoreDNS"

export KUBECONFIG="${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}"
export NODE_ROLE=${NODE_ROLE:-none}
export NODE_TYPE=${NODE_TYPE:-none}
export ETCD_RESTORE=${ETCD_RESTORE:-false}
export INSTALL_MULTINODE=${INSTALL_MULTINODE:-false}
export CERT_CA_FILE_PATH=${CERT_CA_FILE_PATH:-""}
export CERT_TLS_FILE_PATH=${CERT_TLS_FILE_PATH:-""}
export CERT_KEY_FILE_PATH=${CERT_KEY_FILE_PATH:-""}
export ALT_CERT_CA_FILE_PATH=${ALT_CERT_CA_FILE_PATH:-""}
export ALT_CERT_TLS_FILE_PATH=${ALT_CERT_TLS_FILE_PATH:-""}
export ALT_CERT_KEY_FILE_PATH=${ALT_CERT_KEY_FILE_PATH:-""}
export TOKEN_SIGNING_CERT=${TOKEN_SIGNING_CERT:-""}
export ADDITIONAL_CA_CERTS=${ADDITIONAL_CA_CERTS:-""}
export TOKEN_SIGNING_PASS=${TOKEN_SIGNING_PASS:-""}
export INSTALLSTANDALONEAICENTER=${INSTALLSTANDALONEAICENTER:-false}

VERSION=${VERSION:-""}
UPDATE_FQDN=""

export INSTALL_RKE2_CHANNEL="v1.21"
export INSTALL_RKE2_VERSION="v1.21.4+rke2r2"

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/certificate_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/cluster_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/longhornutils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/cilium_utils.sh

#FUNCTIONS

########
# Parse input json file to set the config values
# Globals:
#   INFILE
# Outputs:
#   Sets global config variables
########
function parse_input_file() {
  parse_fqdn_values
  # TODO: remove since this is 90% redundant with parse_fqdn_values

  CLUSTER_PORT="$(get_mandatory_json_value ".cluster_port")"
  VERSION="$(get_optional_json_value ".version")"
  FIXED_RKE_ADDRESS="$(get_mandatory_json_value ".fixed_rke_address")"
  LONGHORN_VERSION="$(get_chart_version_from_manifest "longhorn" "vendor")"
  REPLICAS="$(get_mandatory_json_value ".infra.longhorn.replicas")"
  HA_REPLICAS="$(get_mandatory_json_value ".infra.longhorn.ha_replicas")"
  MOUNT_POINT="$(get_mandatory_json_value ".infra.longhorn.mount")"
  STALE_REPLICA_TIMEOUT="$(get_mandatory_json_value ".infra.longhorn.staleReplicaTimeout")"
  DOCKER_REGISTRY_NODEPORT="$(get_mandatory_json_value ".infra.docker_registry.node_port")"
  DOCKER_REGISTRY_LOCAL_USERNAME="$(get_mandatory_json_value ".infra.docker_registry.username")"
  DOCKER_REGISTRY_LOCAL_PASSWORD="$(get_mandatory_json_value ".infra.docker_registry.password")"
  DOCKER_REGISTRY_URL="$(get_mandatory_json_value ".registries.docker.url")"
  DOCKER_REGISTRY_PULL_SECRET="$(get_optional_json_value ".registries.docker.pull_secret_value")"
  DOCKER_REGISTRY_USERNAME="$(get_optional_json_value ".registries.docker.username")"
  DOCKER_REGISTRY_PASSWORD="$(get_optional_json_value ".registries.docker.password")"
  RKE_TOKEN="$(get_mandatory_json_value ".rke_token")"
  CLOUD_TEMPLATE_VENDOR="$(get_optional_json_value ".cloud_template_vendor")"
  CLOUD_TEMPLATE_SOURCE="$(get_optional_json_value ".cloud_template_source")"
  BACKUP_INTERVAL="$(get_mandatory_json_value ".backup_interval")"
  STORAGE_CLASS="$(get_mandatory_json_value ".storage_class")"
  SYSTEM_RESERVED_CPU="$(get_mandatory_json_value ".kubelet.system_reserved_cpu")"
  SYSTEM_RESERVED_MEMORY="$(get_mandatory_json_value ".kubelet.system_reserved_memory")"
  SYSTEM_RESERVED_EPHEMERAL_STORAGE="$(get_mandatory_json_value ".kubelet.system_reserved_ephemeral_storage")"
  KUBE_RESERVED_CPU="$(get_mandatory_json_value ".kubelet.kube_reserved_cpu")"
  KUBE_RESERVED_MEMORY="$(get_mandatory_json_value ".kubelet.kube_reserved_memory")"
  KUBE_RESERVED_EPHEMERAL_STORAGE="$(get_mandatory_json_value ".kubelet.kube_reserved_ephemeral_storage")"
  EVICTION_HARD_MEMORY_AVAILABLE="$(get_mandatory_json_value ".kubelet.eviction_hard_memory_available")"
  EVICTION_HARD_NODEFS_AVAILABLE="$(get_mandatory_json_value ".kubelet.eviction_hard_nodefs_available")"
  IMAGE_GC_HIGH_THRESHOLD="$(get_mandatory_json_value ".kubelet.image_gc_high_threshold")"
  IMAGE_GC_LOW_THRESHOLD="$(get_mandatory_json_value ".kubelet.image_gc_low_threshold")"
  NODE_STATUS_UPDATE_FREQUENCY="$(get_mandatory_json_value ".kubelet.node_status_update_frequency")"
  DEFAULT_NOT_READY_TOLERATION_SECONDS="$(get_mandatory_json_value ".kubelet.default_not_ready_toleration_seconds")"
  DEFAULT_UNREACHABLE_TOLERATION_SECONDS="$(get_mandatory_json_value ".kubelet.default_unreachable_toleration_seconds")"
  NODE_MONITOR_PERIOD="$(get_mandatory_json_value ".kubelet.node_monitor_period")"
  NODE_MONITOR_GRACE_PERIOD="$(get_mandatory_json_value ".kubelet.node_monitor_grace_period")"
  PROXY_ENABLED="$(get_optional_json_value ".proxy.enabled")"
  HTTP_PROXY="$(get_optional_json_value ".proxy.http_proxy")"
  HTTPS_PROXY="$(get_optional_json_value ".proxy.https_proxy")"
  NO_PROXY="$(get_optional_json_value ".proxy.no_proxy")"
  PROFILE="$(get_mandatory_json_value ".profile")"
  CUSTOM_DNS_RESOLVER="$(get_optional_json_value ".infra.custom_dns_resolver")"
  PODMAN_FOLDER="$(get_mandatory_json_value ".infra.offline.dir")/lib/containers"

  TELEMETRY_OPT_OUT="$(get_optional_json_value ".telemetry_optout")"

  ADDITIONAL_CA_CERTS="$(get_optional_json_value ".additional_ca_certs")"
  ADDITIONAL_CA_CERTS="${ADDITIONAL_CA_CERTS/#\~/$HOME}"

  RKE2_CALICO_MTU_LIMIT="$(get_json_value_or_return_default_value ".rke2.calico_mtu_limit" 1350)"
  ZONE_RESILIENCE="$(get_json_value_or_return_default_value ".zone_resilience" "false")"
  LH_RECURRING_SNAPSHOT_SCHEDULE="$(get_mandatory_json_value ".infra.longhorn.recurring_snapshot_schedule")"

  ACTIVE_ACTIVE="$(get_optional_json_value ".active_active.enabled")"
  PRIMARY_CLUSTER="$(get_optional_json_value ".active_active.primary")"

  primary_cluster_kubeconfig_base64="$(get_optional_json_value ".active_active.primary_kube_config")"
  PRIMARY_CLUSTER_KUBECONFIG=$(echo "${primary_cluster_kubeconfig_base64}"  | tr -d '"' | base64 --decode || true)
  PRIMARY_CLUSTER_FQDN=$(echo "${PRIMARY_CLUSTER_KUBECONFIG}" | grep -E 'server:' | cut -d'/' -f 3 | cut -d':' -f1)

  # Disable telemetry in case of offline
  if [[ "${INSTALL_TYPE}" == "offline" ]];
  then
    TELEMETRY_OPT_OUT="true"
  fi
}

function get_http_return_code() {
  local url="$1"
  local return_code

  return_code=$(curl -m 30 -k -s "${url}" -o /dev/null -w "%{http_code}")
  echo "${return_code}"
}

function validate_primary_connectivity() {
  local error_messages=()
  local return_code

  return_code=$(get_http_return_code "https://${FIXED_RKE_ADDRESS}:9345/cacerts")
  [[ $return_code -ne 200 ]] && error_messages+=("URL: 'https://${FIXED_RKE_ADDRESS}:9345/cacerts' returned unexpected http code: '${return_code}' Expected: '200' Ensure network allows connectivity on port: 9345 and RKE2 server service is running on already configured server(s)")

  return_code=$(get_http_return_code "https://${FIXED_RKE_ADDRESS}:6443")
  [[ $return_code -ne 401 ]] && error_messages+=("URL: 'https://${FIXED_RKE_ADDRESS}:6443' returned unexpected http code: '${return_code}' Expected: '401' Ensure network allows connectivity on port: 6443 and RKE2 server service is running on already configured server(s)")

  if [[ ${#error_messages[@]} != 0 ]] ; then
    failure_output "Primary Connectivity Validation Failed" "${error_messages[@]}"
  fi
}

## Function to populate zone&region labels
function populate_region_zone_node_labels(){
  local node_region
  local node_zone
  local region_zone_separator_key="-"

  detect_machine_provider
  case "${provider}" in
    azure)
      region_cmd='curl -m 5 -s -H "Metadata:true" "http://169.254.169.254/metadata/instance/compute/location?api-version=2017-12-01&format=text"'
      zone_cmd='curl -m 5 -s -H "Metadata:true" "http://169.254.169.254/metadata/instance/compute/zone?api-version=2017-12-01&format=text"'
    ;;
    amazon)
      region_cmd='curl -m 5 -s "http://169.254.169.254/latest/meta-data/placement/region"'
      zone_cmd='curl -m 5 -s "http://169.254.169.254/latest/meta-data/placement/availability-zone"'
      region_zone_separator_key=""
    ;;
    google)
      region_cmd="curl -m 5 -s 'http://169.254.169.254/computeMetadata/v1/instance/zone'  -H 'Metadata-Flavor: Google' | awk -F'/' '{print \$NF}' | rev | cut -d'-' -f 2- | rev"
      zone_cmd="curl -m 5 -s 'http://169.254.169.254/computeMetadata/v1/instance/zone'  -H 'Metadata-Flavor: Google' | awk -F'/' '{print \$NF}'"
      region_zone_separator_key=""
    ;;
    unknown)
      ## Use "dc" as default region due to ceph issue with region="default"
      ## https://github.com/rook/rook/issues/4993
      region_cmd='echo -n "dc"'
      zone_cmd='echo -n "1"'
    ;;
  esac

  node_region=$(eval "${region_cmd}")
  node_zone=$(eval "${zone_cmd}")

  ## Use default region "dc" if empty or null
  if [[ -z ${node_region} ]] || [[ ${node_region} == "null" ]] || echo "${node_region}" | grep -i -q 'not found'
  then
    [[ ${provider} != "unknown" ]] && warn "Could not determine region of the machine when provider is ${provider}, using default region 'dc'"
    node_region="dc"
  fi

  if [[ -z ${node_zone} ]] || [[ ${node_zone} == "null" ]] || echo "${node_zone}" | grep -i -q 'not found'
  then
    [[ ${provider} != "unknown" ]] && warn "Could not determine zone of the machine when provider is ${provider}, using '1'"
    node_zone="1"
  fi

  ## Remove region prefix from zone
  node_zone=${node_zone#"$node_region"}

  NODE_LABELS_REGION_ZONE="topology.kubernetes.io/region=${node_region},topology.kubernetes.io/zone=${node_region}${region_zone_separator_key}${node_zone}"
}

function detect_machine_provider(){
  ##try azure
  if curl -o "${dev_null}" -f -m 5 -s -H "Metadata:true" "http://169.254.169.254/metadata/instance?api-version=2017-12-01"
  then
    provider=azure
  elif curl -o "${dev_null}" -f -m 5 -s "http://169.254.169.254/latest/dynamic/instance-identity/document"
  then
    provider=amazon
  elif curl -o "${dev_null}" -f -m 5 -s -H "Metadata-Flavor: Google" "http://169.254.169.254/computeMetadata/v1/"
  then
    provider=google
  else
    provider=unknown
  fi
}

function populate_node_labels(){
  local EXTRA_K8S_NODE_LABELS=""
  populate_region_zone_node_labels

  RKE_NODE_LABELS="${NODE_LABELS_REGION_ZONE}"

  # shellcheck disable=SC1091
  [[ -f /etc/default/k8s-node-labels ]] && source /etc/default/k8s-node-labels
  if [[ -n ${EXTRA_K8S_NODE_LABELS} ]]
  then
    RKE_NODE_LABELS="${NODE_LABELS_REGION_ZONE},${EXTRA_K8S_NODE_LABELS}"
  fi
}

#RKE2 FUNCTIONS
function generate_rke2_config(){
  local config_path="/etc/rancher/rke2/config.yaml"
  local kube_scheduler_config_filepath="/etc/rancher/rke2/kube-scheduler/config.yaml"
  local kube_scheduler_kubeconfig_filepath="/var/lib/rancher/rke2/server/cred/scheduler.kubeconfig"
  local EXISTING_TOKEN=""
  local kube_scheduler_config_folderpath
  local kube_scheduler_config_filename
  [[ -f ${config_path} ]] && EXISTING_TOKEN=$(grep -m 1 token ${config_path} | awk -F ':' '{print $2}' | xargs)
  USE_CILIUM=1
  [[ -f ${config_path} ]] && USE_CILIUM=$(grep -c cilium ${config_path} || true)
  [[ -n "${EXISTING_TOKEN}" && "${EXISTING_TOKEN}" != "${RKE_TOKEN}" ]] && error_with_code "11026"
  local is_selinux_enabled
  mkdir -p /etc/rancher/rke2/
  kube_scheduler_config_folderpath=$(dirname "${kube_scheduler_config_filepath}")
  kube_scheduler_config_filename=$(basename "${kube_scheduler_config_filepath}")
  mkdir -p "${kube_scheduler_config_folderpath}"
  cat > "${kube_scheduler_config_filepath}" <<EOF
apiVersion: kubescheduler.config.k8s.io/v1beta1
kind: KubeSchedulerConfiguration
clientConnection:
  kubeconfig: ${kube_scheduler_kubeconfig_filepath}
profiles:
  - pluginConfig:
    - name: PodTopologySpread
      args:
        defaultingType: List
        defaultConstraints:
        - maxSkew: 1
          topologyKey: "topology.kubernetes.io/zone"
          whenUnsatisfiable: ScheduleAnyway
        - maxSkew: 1
          topologyKey: "kubernetes.io/hostname"
          whenUnsatisfiable: ScheduleAnyway
    plugins:
      score:
        disabled:
        - name: PodTopologySpread
        enabled:
        - name: PodTopologySpread
          weight: 500
EOF
  is_selinux_enabled="false"
  #selinuxenabled && is_selinux_enabled="true"
  # shellcheck disable=SC2001
  #common to both agent and server
  #https://docs.rke2.io/install/install_options/agent_config.html
  #https://docs.rke2.io/install/install_options/server_config.html
  #https://github.com/rancher/rancher/issues/29997#issuecomment-725261582
  cat >> "${config_path}" <<EOF
token: ${RKE_TOKEN}
selinux: ${is_selinux_enabled}
kubelet-arg:
- "max-pods=500"
- "enforce-node-allocatable=pods"
- "system-reserved=cpu=${SYSTEM_RESERVED_CPU},memory=${SYSTEM_RESERVED_MEMORY},ephemeral-storage=${SYSTEM_RESERVED_EPHEMERAL_STORAGE}"
- "kube-reserved=cpu=${KUBE_RESERVED_CPU},memory=${KUBE_RESERVED_MEMORY},ephemeral-storage=${KUBE_RESERVED_EPHEMERAL_STORAGE}"
- "eviction-hard=memory.available<${EVICTION_HARD_MEMORY_AVAILABLE},nodefs.available<${EVICTION_HARD_NODEFS_AVAILABLE}"
- "image-gc-high-threshold=${IMAGE_GC_HIGH_THRESHOLD}"
- "image-gc-low-threshold=${IMAGE_GC_LOW_THRESHOLD}"
- "node-status-update-frequency=${NODE_STATUS_UPDATE_FREQUENCY}"
- "node-labels=${RKE_NODE_LABELS}"
EOF
  # This has to be appended to kubelet-arg section in config file, so DO NOT separate it from the above block
  [[ -n "${CUSTOM_DNS_RESOLVER}" ]] && echo "- \"resolv-conf=${CUSTOM_DNS_RESOLVER}\"" >> "${config_path}"

  # shellcheck disable=SC2001
  # props specific to server
  [[ ${NODE_ROLE} != "agent" && "${USE_CILIUM}" -gt 0 ]] && cat >> "${config_path}" <<EOF
cni: cilium
EOF
  [[ ${NODE_ROLE} != "agent" ]] && cat >> "${config_path}" <<EOF
write-kubeconfig-mode: "0600"
etcd-snapshot-retention: 7
etcd-snapshot-schedule-cron: "*/${BACKUP_INTERVAL} * * * *"
tls-san:
$(printf -- '- %s\n' "${FIXED_RKE_ADDRESS}")
- kubernetes.default.svc
disable:
- rke2-ingress-nginx
kube-apiserver-arg:
- "feature-gates=TTLAfterFinished=true"
- "default-not-ready-toleration-seconds=${DEFAULT_NOT_READY_TOLERATION_SECONDS}"
- "default-unreachable-toleration-seconds=${DEFAULT_UNREACHABLE_TOLERATION_SECONDS}"
kube-controller-manager-arg:
- "node-cidr-mask-size=22"
- "pod-eviction-timeout=10s"
- "feature-gates=TTLAfterFinished=true"
- "node-monitor-period=${NODE_MONITOR_PERIOD}"
- "node-monitor-grace-period=${NODE_MONITOR_GRACE_PERIOD}"
disable-cloud-controller: true
control-plane-resource-requests:
- "kube-apiserver-cpu=200m,kube-apiserver-memory=512M,kube-scheduler-cpu=50m,kube-scheduler-memory=50M,etcd-cpu=150m,etcd-memory=300M,kube-controller-manager-cpu=50m,kube-controller-manager-memory=200M,kube-proxy-cpu=100m,kube-proxy-memory=100M"
kube-scheduler-extra-mount: ${kube_scheduler_config_folderpath}:/kube-scheduler:ro
kube-scheduler-arg: "--config=/kube-scheduler/${kube_scheduler_config_filename}"
EOF

  if [[ ${NODE_ROLE} != "none" ]]; then
    validate_primary_connectivity
    cat >> "${config_path}" <<< "server: https://${FIXED_RKE_ADDRESS}:9345"
  fi

  [[ ${NODE_ROLE} == "agent" && ${NODE_TYPE} != "agent" ]] && add_node_taint_config "${config_path}"

  if [[ "$(service_exists NetworkManager)" == "true" ]];
  then
    #https://docs.rke2.io/known_issues/#networkmanager
    local canalConfigPath="/etc/NetworkManager/conf.d/rke2-canal.conf"
    local networkDir="/etc/NetworkManager/conf.d/"
    if [ -d "${networkDir}" ]; then
      cat << EOF > "${canalConfigPath}"
[keyfile]
unmanaged-devices=interface-name:cali*;interface-name:flannel*
EOF
    
      systemctl reload NetworkManager || info "Skipping reload NetworkManager because service is disabled"
    fi
  fi

  local registriesPath="/etc/rancher/rke2/registries.yaml"
  if [[ ${INSTALL_TYPE} == "offline" ]]; then
    cat > "${registriesPath}" <<EOF
mirrors:
  docker-registry.docker-registry.svc.cluster.local:5000:
    endpoint:
      - "https://${DOCKER_REGISTRY_URL}"
  docker.io:
    endpoint:
      - "https://${DOCKER_REGISTRY_URL}"
  ${DOCKER_REGISTRY_URL}:
    endpoint:
      - "https://${DOCKER_REGISTRY_URL}"
configs:
  "localhost:${DOCKER_REGISTRY_NODEPORT}":
    tls:
      insecure_skip_verify: true
    auth:
      username: ${DOCKER_REGISTRY_LOCAL_USERNAME}
      password: ${DOCKER_REGISTRY_LOCAL_PASSWORD}
EOF
  else # INSTALL_TYPE == "online"
    cat >> "${registriesPath}" <<EOF
mirrors:
  ${DOCKER_REGISTRY_URL}:
    endpoint:
      - "https://${DOCKER_REGISTRY_URL}"
  docker.io:
    endpoint:
      - "https://${DOCKER_REGISTRY_URL}"
  index.docker.io:
    endpoint:
      - "https://${DOCKER_REGISTRY_URL}"
EOF
    if [[ -n ${DOCKER_REGISTRY_USERNAME} ]] && [[ -n ${DOCKER_REGISTRY_PASSWORD} ]]; then
      cat >> "${registriesPath}" <<EOF
configs:
  ${DOCKER_REGISTRY_URL}:
    auth:
      username: ${DOCKER_REGISTRY_USERNAME}
      password: ${DOCKER_REGISTRY_PASSWORD}
EOF
    else
      cat >> "${registriesPath}" <<EOF
configs:
  ${DOCKER_REGISTRY_URL}:
    auth: {}
EOF
    fi
  fi

# Add proxy configuration
if [[ "${PROXY_ENABLED}" == "true" ]] ; then
  local rke2ProxyConfigurationPath="/etc/sysconfig/rke2-server"
  if [[ ${NODE_ROLE} == "agent" ]]; then
    rke2ProxyConfigurationPath="/etc/sysconfig/rke2-agent"
  fi
  local is_selinux_enabled
  is_selinux_enabled="false"
  #selinuxenabled && is_selinux_enabled="true"
  cat > "${rke2ProxyConfigurationPath}" <<EOF
HOME=/root
RKE2_SELINUX=${is_selinux_enabled}
https_proxy=${HTTPS_PROXY}
http_proxy=${HTTP_PROXY}
no_proxy=${NO_PROXY},127.0.0.0/8,localhost,kubernetes,.svc,.local,.cluster
EOF

 [[ -f ${rke2ProxyConfigurationPath} ]] || {
    error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_K8S_INSTALL_MESSAGE}";
    error_with_code "11001";
  }
fi

# Add rke2 canal configuration
if [[ ${NODE_ROLE} != "agent" ]] ; then
  local rke2CanalConfigurationDir="/var/lib/rancher/rke2/server/manifests/"
  local rke2CanalConfigurationPath="/var/lib/rancher/rke2/server/manifests/rke2-canal-mtu-resize.yaml"

  mkdir -p "${rke2CanalConfigurationDir}"

  cat > "${rke2CanalConfigurationPath}" <<EOF
apiVersion: helm.cattle.io/v1
kind: HelmChartConfig
metadata:
  name: rke2-canal
  namespace: kube-system
spec:
  valuesContent: |-
    calico:
      vethuMTU: "${RKE2_CALICO_MTU_LIMIT}"
EOF

  configure_cilium

 [[ -f ${rke2CanalConfigurationPath} ]] || {
    error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_K8S_INSTALL_MESSAGE}";
    error_with_code "11027";
  }
fi

  [[ -f ${registriesPath} ]] || {
    error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_K8S_INSTALL_MESSAGE}";
    error_with_code "11002";
  }

  [[ -f ${config_path} ]] || {
    error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_K8S_INSTALL_MESSAGE}";
    error_with_code "11003";
  }
}

# GPU related modifications to the containerd
function configure_gpu() {
  # Check if nvidia drivers and runtime are installed before making changes to containerd config
  if ! nvidia-smi &>>"${dev_null}";
  then
    warn "GPU Drivers are not installed on the VM. Please refer the documentation."
    return
  fi
  if ! which nvidia-container-runtime &>>"${dev_null}";
  then
    warn "Nvidia container runtime is not installed on the VM. Please refer the documentation."
    return
  fi

  grep "nvidia-container-runtime" /var/lib/rancher/rke2/agent/etc/containerd/config.toml &>>"${dev_null}" && info "GPU containerd changes already applied" && return

  awk '1;/plugins.cri.containerd]/{print "  default_runtime_name = \"nvidia-container-runtime\""}' /var/lib/rancher/rke2/agent/etc/containerd/config.toml > /var/lib/rancher/rke2/agent/etc/containerd/config.toml.tmpl
  echo -e '\n[plugins.linux]\n  runtime = "nvidia-container-runtime"' >> /var/lib/rancher/rke2/agent/etc/containerd/config.toml.tmpl
  echo -e '\n[plugins.cri.containerd.runtimes.nvidia-container-runtime]\n  runtime_type = "io.containerd.runc.v2"\n  [plugins.cri.containerd.runtimes.nvidia-container-runtime.options]\n    BinaryName = "nvidia-container-runtime"' >> /var/lib/rancher/rke2/agent/etc/containerd/config.toml.tmpl

  if [[ ${NODE_ROLE} == "agent" ]]; then
    systemctl restart rke2-agent.service || {
      error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_GPU_SETUP_MESSAGE}";
      error_with_code "11004";
    }
  else
    systemctl restart rke2-server.service || {
      error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_GPU_SETUP_MESSAGE}";
      error_with_code "11004";
    }
  fi
}

function enable_log_persistence() {
  #https://man7.org/linux/man-pages/man5/journald.conf.5.html
  #if /var/log/journal folder does not exist, then logs that correspond to systemd services(ex: rke2-server) are volatile and are stored under /run/log/journal folder
  mkdir -p /var/log/journal

  if test -f "/etc/systemd/journald.conf"
  then
    add_or_replace_option SystemMaxUse 1G /etc/systemd/journald.conf
  fi
}

function configure_auditd_service() {
  #https://github.com/osquery/osquery/issues/3148
  if systemctl is-enabled auditd &>>"${dev_null}";
  then
    # enabled set to 2 means, it is immutable
    if [[ "$(get_auditctl_parameter_value "enabled")" == "2" ]]; then
      if [[ "$(get_auditctl_parameter_value "backlog_wait_time")" != "1" ]]; then
        error "Automation Suite require backlog_wait_time to be set 1. Please follow our troubleshooting guide to set backlog_wait_time to 1.\n Troubleshooting Guide Link - https://docs.uipath.com/automation-suite/docs/troubleshooting"
      fi
    elif test -f "/etc/audit/rules.d/audit.rules"; then
      # Perform following steps only in case of mutable auditd
      auditctl --backlog_wait_time 1
      sed -i -e '/--backlog_wait_time/ s/ .*/ 1/' /etc/audit/rules.d/audit.rules
      service auditd stop &>>"${dev_null}"
      service auditd start &>>"${dev_null}"
    fi
  fi
}

#systemd service to drain the k8s node before node shutdown
function enable_graceful_node_shutdown() {
  local SERVICE_NAME

  #shellcheck disable=SC2216
  yes | cp "${MODULES_DIR}"/node-drain.sh /opt/node-drain.sh  &>>"${dev_null}"
  chmod +x /opt/node-drain.sh

  if [[ ${NODE_ROLE} == "agent" ]]; then
    SERVICE_NAME=rke2-agent.service
  else
    SERVICE_NAME=rke2-server.service
    kubectl apply -f "${SCRIPT_DIR}/k8s_manifests/node_drain_rbac.yaml"
  fi

  cat > /etc/systemd/system/node-drain.service <<EOF
[Unit]
Description=draining the node to make sure pods terminate gracefully before kubelet shuts down
After=${SERVICE_NAME}

[Service]
Type=oneshot
ExecStart=/bin/true
ExecStop=/bin/bash /opt/node-drain.sh
RemainAfterExit=true
TimeoutStopSec=300

[Install]
WantedBy=multi-user.target
EOF

  cat > /etc/systemd/system/node-uncordon.service <<EOF
[Unit]
Description=Automation Suite Node Start Env Setup
After=${SERVICE_NAME}
StartLimitIntervalSec=2000
StartLimitBurst=5

[Service]
Restart=on-failure
RestartSec=300
ExecStart=/bin/bash /opt/node-drain.sh nodestart

[Install]
WantedBy=multi-user.target
EOF
  
  #shellcheck disable=SC2129
  systemctl start node-drain.service &>>"${dev_null}"
  systemctl enable node-drain.service &>>"${dev_null}"

  systemctl start node-uncordon.service &>>"${dev_null}"
  systemctl enable node-uncordon.service &>>"${dev_null}"

  systemctl daemon-reload
}


function install_rke2() {
  enable_log_persistence
  configure_auditd_service

  if systemctl is-enabled firewalld &>>"${dev_null}";
  then
    systemctl stop firewalld
    systemctl disable firewalld
  fi

  [[ ${NODE_ROLE} == "agent" ]] && export INSTALL_RKE2_TYPE="agent"
  ./rke2/rke2.sh

  ## Check if selinux enabled.
  is_selinux_enabled=false

  if [ "$is_selinux_enabled" = true ]
  then
    ## Load kernel modules
    ## Required for istio side car to work
    modprobe br_netfilter
    modprobe xt_REDIRECT
    modprobe xt_owner
    echo -e 'br_netfilter\nxt_REDIRECT\nxt_owner' > /etc/modules-load.d/istio-iptables.conf

    ## Allow pods to use hostpath used by ceph cluster
    ## Required for ceph component to use the PVC
    mkdir -p /var/lib/rook
    chcon -v -R -t container_file_t /var/lib/rook

    ## Verify the rke2 selinux module
    if ! semodule -l | grep -q rke2
    then
      warn "Required selinux module: rke2 not loaded, this may lead to unexpected behavior"
    fi
  fi

  local SERVICE_NAME

  if [[ ${NODE_ROLE} == "agent" ]]; then
    SERVICE_NAME=rke2-agent.service
  else
    SERVICE_NAME=rke2-server.service
  fi

  systemctl enable ${SERVICE_NAME}
  if [[ ${DO_UPGRADE} == "1" ]]; then
    systemctl restart ${SERVICE_NAME}
  fi

  local try=0
  local maxtry=5
  local status="notready"
  while [[ ${status} == "notready" ]]  && (( try != maxtry )) ; do
    try=$((try+1))
    info "Trying to start the rke2 service...${try}/${maxtry}";
    timeout --foreground 1200 systemctl start ${SERVICE_NAME} && status="ready"

    if [[ "$?" != "124" ]]; then
      sleep 120
    elif (( try >=  2 )); then
      break
    fi

    [[ "${status}" == "ready" ]] || systemctl stop ${SERVICE_NAME}

    [[ "${status}" == "ready" ]] && {
      systemctl daemon-reload
      ETCD_MOUNT_SERVICE="var-lib-rancher-rke2-server-db.mount"
      VAR_RANCHER_MOUNT_SERVICE="var-lib-rancher.mount"
      VAR_KUBELET_MOUNT_SERVICE="var-lib-kubelet.mount"
      MOUNT_DEPENDENCIES=""
      systemctl is-enabled $ETCD_MOUNT_SERVICE -q 2>>"${dev_null}" && MOUNT_DEPENDENCIES="$ETCD_MOUNT_SERVICE"
      systemctl is-enabled $VAR_RANCHER_MOUNT_SERVICE -q 2>>"${dev_null}" && MOUNT_DEPENDENCIES="$MOUNT_DEPENDENCIES $VAR_RANCHER_MOUNT_SERVICE"
      systemctl is-enabled $VAR_KUBELET_MOUNT_SERVICE -q 2>>"${dev_null}" && MOUNT_DEPENDENCIES="$MOUNT_DEPENDENCIES $VAR_KUBELET_MOUNT_SERVICE"
      [[ -n "${MOUNT_DEPENDENCIES}" ]] && {
        info "Adding mount services dependency on RKE2 service : ${MOUNT_DEPENDENCIES}"
        mkdir -p /etc/systemd/system/${SERVICE_NAME}.d/
        cat > /etc/systemd/system/${SERVICE_NAME}.d/custom.conf <<EOF
[Unit]
After=${MOUNT_DEPENDENCIES}
Requires=${MOUNT_DEPENDENCIES}
EOF
        systemctl daemon-reload
      }
    }
  done

  [[ ${status} == "ready" ]] || {
    error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_K8S_INSTALL_MESSAGE}";
    error_with_code "11005";
  }
}

function longhorn_prereq(){
  info "Validate mount propagation..."
  local try=0
  local maxtry=30

  # if INSTALL_TYPE is offline, then the registry image is already present on the machine
  # if INSTALL_TYPE is online, then images need to be pulled from DOCKER_REGISTRY_URL
  if [[ ${INSTALL_TYPE} == "online" ]]; then
    sed -i "s;##REGISTRY_PLACEHOLDER##;${DOCKER_REGISTRY_URL}/;" ./longhorn/tests/longhorn_test_ds.yaml
    sed -i "s;##REGISTRY_PLACEHOLDER##;${DOCKER_REGISTRY_URL}/;" ./longhorn/tests/pod_with_pvc.yaml
  else
    sed -i "s;##REGISTRY_PLACEHOLDER##;;" ./longhorn/tests/longhorn_test_ds.yaml
    sed -i "s;##REGISTRY_PLACEHOLDER##;;" ./longhorn/tests/pod_with_pvc.yaml
  fi

  kubectl create namespace longhorn-system --dry-run=client -o yaml | kubectl apply -f - &>>"${dev_null}"
  while ! kubectl get namespaces | grep -q "longhorn-system" && (( try != maxtry )) ; do
    try=$((try+1))
    info "Waiting for 'longhorn-system' namespace to be ready...${try}/${maxtry}" &&  sleep 10;
  done

  kubectl get namespaces | grep -q "longhorn-system" || error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_STORAGE_SETUP_MESSAGE}"
  kubectl get namespaces | grep -q "longhorn-system" || error_with_code "11006"

  # Label longhorn-system namespace, this is required in longhorn networkpolicy
  kubectl label namespace longhorn-system name=longhorn-system &>>"${dev_null}" || true

  MOUNTED=$(./longhorn/environment_check.sh \
    | grep 'MountPropagation is enabled!')

  [[ -n ${MOUNTED} ]] || {
    error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_STORAGE_SETUP_MESSAGE}";
    error_with_code "11007";
  }
  info "Validation complete..."
}

function patch_existing_volumes_stale_timeout() {
  # patch existing volume objects with new stale replica timeout
  volume_objects=$(kubectl -n longhorn-system get volumes.longhorn.io --no-headers -o custom-columns=Name:.metadata.name)
  for vol in $volume_objects;
  do
    kubectl -n longhorn-system patch volumes.longhorn.io "$vol" --type merge -p "{\"spec\" : { \"staleReplicaTimeout\": ${STALE_REPLICA_TIMEOUT} } }" || true
  done
}

function create_longhorn_stale_folder_cleanup_cronjob() {
  cp ./longhorn/longhorn-replica-folder-cleanup.yaml ./longhorn/longhorn-replica-folder-cleanup-tmp.yaml
  # if INSTALL_TYPE is offline, then the registry image is already present on the machine
  # if INSTALL_TYPE is online, then images need to be pulled from DOCKER_REGISTRY_URL
  if [[ ${INSTALL_TYPE} == "online" ]]; then
    sed -i "s;##REGISTRY_PLACEHOLDER##;${DOCKER_REGISTRY_URL}/;" ./longhorn/longhorn-replica-folder-cleanup-tmp.yaml
  else
    sed -i "s;##REGISTRY_PLACEHOLDER##;;" ./longhorn/longhorn-replica-folder-cleanup-tmp.yaml
  fi
  sed -i "s;##LONGHORN_DATADISK_PLACEHOLDER##;${MOUNT_POINT}/;" ./longhorn/longhorn-replica-folder-cleanup-tmp.yaml
  kubectl apply -f ./longhorn/longhorn-replica-folder-cleanup-tmp.yaml 1>>"${dev_null}" || true
}

function create_longhorn_loop_device_cleaner_daemonset() {
  cp ./longhorn/longhorn-loop-device-cleaner.yaml ./longhorn/longhorn-loop-device-cleaner-tmp.yaml
  # if INSTALL_TYPE is offline, then the registry image is already present on the machine
  # if INSTALL_TYPE is online, then images need to be pulled from DOCKER_REGISTRY_URL
  if [[ ${INSTALL_TYPE} == "online" ]]; then
    sed -i "s;##REGISTRY_PLACEHOLDER##;${DOCKER_REGISTRY_URL}/;" ./longhorn/longhorn-loop-device-cleaner-tmp.yaml
  else
    sed -i "s;##REGISTRY_PLACEHOLDER##;;" ./longhorn/longhorn-loop-device-cleaner-tmp.yaml
  fi
  kubectl apply -f ./longhorn/longhorn-loop-device-cleaner-tmp.yaml 1>>"${dev_null}" || true
}

function install_longhorn(){

  declare -a LONGHORN_ARGS
  [[ ${INSTALL_TYPE} == "online" ]] && LONGHORN_ARGS+=(--set global.cattle.systemDefaultRegistry="${DOCKER_REGISTRY_URL}")
  [[ -n ${LONGHORN_VERSION} ]] && LONGHORN_ARGS+=(--version="${LONGHORN_VERSION}")

  LONGHORN_ARGS+=(--set defaultSettings.autoDeletePodWhenVolumeDetachedUnexpectedly=true)
  if [[ $PROFILE != "ha" ]]; then
    LONGHORN_ARGS+=(--set defaultSettings.guaranteedReplicaManagerCPU=1)
    LONGHORN_ARGS+=(--set defaultSettings.guaranteedEngineManagerCPU=1)
  else
    LONGHORN_ARGS+=(--set defaultSettings.guaranteedReplicaManagerCPU=10)
    LONGHORN_ARGS+=(--set defaultSettings.guaranteedEngineManagerCPU=10)
  fi;
  LONGHORN_ARGS+=(--set defaultSettings.storageMinimalAvailablePercentage=5)
  LONGHORN_ARGS+=(--set defaultSettings.allowNodeDrainWithLastHealthyReplica=false)
  LONGHORN_ARGS+=(--set defaultSettings.defaultDataLocality=best-effort)
  LONGHORN_ARGS+=(--set defaultSettings.nodeDownPodDeletionPolicy=delete-both-statefulset-and-deployment-pod)

  LONGHORN_ARGS+=(--set defaultSettings.priorityClass=system-cluster-critical)
  LONGHORN_ARGS+=(--set longhornManager.priorityClass=system-cluster-critical)
  LONGHORN_ARGS+=(--set longhornDriver.priorityClass=system-cluster-critical)
  LONGHORN_ARGS+=(--set longhornUI.priorityClass=system-cluster-critical)

  if kubectl -n longhorn-system get cm longhorn-default-setting &>>"${dev_null}";
  then
    OVERPROVISIONING_PERCENTAGE=$(kubectl -n longhorn-system get cm longhorn-default-setting -o jsonpath='{.data.default-setting\.yaml}' | grep storage-over-provisioning-percentage: | cut -d ':' -f2 | xargs)
    if [[ -n "${OVERPROVISIONING_PERCENTAGE}" ]];
    then
      info "Longhorn default settings found with storageOverProvisioningPercentage set to ${OVERPROVISIONING_PERCENTAGE}"
      LONGHORN_ARGS+=(--set defaultSettings.storageOverProvisioningPercentage="${OVERPROVISIONING_PERCENTAGE}")
    fi
  else
    info "Longhorn default settings not found, setting storageOverProvisioningPercentage to 100%"
    LONGHORN_ARGS+=(--set defaultSettings.storageOverProvisioningPercentage=100)
  fi

  REPLICAS_BACKUP_ONE=1
  if [[ $INSTALL_MULTINODE == "true" ]]; then
    REPLICAS=${HA_REPLICAS}
    REPLICAS_BACKUP_ONE=2
    LONGHORN_ARGS+=(--set defaultSettings.replicaAutoBalance=best-effort)
    if [[ $ZONE_RESILIENCE == "true" ]]; then
      LONGHORN_ARGS+=(--set defaultSettings.replicaZoneSoftAntiAffinity=false)
    fi
  else
    # reset csi pods to 1 replica for single node
    LONGHORN_ARGS+=(--set csi.attacherReplicaCount=1)
    LONGHORN_ARGS+=(--set csi.provisionerReplicaCount=1)
    LONGHORN_ARGS+=(--set csi.resizerReplicaCount=1)
    LONGHORN_ARGS+=(--set csi.snapshotterReplicaCount=1)
  fi

  LONGHORN_ARGS+=(--set persistence.defaultClassReplicaCount="1")
  LONGHORN_ARGS+=(--set defaultSettings.defaultReplicaCount="1")
  #defaultDataPath can only be set at installation time, not upgrade
  [[ -n ${MOUNT_POINT} ]] && LONGHORN_ARGS+=(--set defaultSettings.defaultDataPath="${MOUNT_POINT}")

  info "Installing longhorn with arguments ${LONGHORN_ARGS[*]}"

  helm upgrade --install --atomic -f ./longhorn/longhorn-tolerations-values.yaml longhorn "${TMP_HELM_CHARTS_DIR}"/longhorn/ \
    --wait --timeout 5m \
    --namespace longhorn-system \
    --create-namespace \
    "${LONGHORN_ARGS[@]}" &>>"${dev_null}" || {
      error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_STORAGE_SETUP_MESSAGE}";
      error_with_code "11008";
    }


  # create default backup storage class
  rm -f ./longhorn/longhorn-storage-backup-tmp.yaml &>>"${dev_null}"
  cp ./longhorn/longhorn-storage-backup.yaml ./longhorn/longhorn-storage-backup-tmp.yaml
  sed -i "s/REPLICA_PLACEHOLDER/${REPLICAS}/" ./longhorn/longhorn-storage-backup-tmp.yaml
  sed -i "s/STALE_REPLICA_TIMEOUT_PLACEHOLDER/${STALE_REPLICA_TIMEOUT}/" ./longhorn/longhorn-storage-backup-tmp.yaml
  # Delete older storageclass definitions so that new parameters are applied
  kubectl delete -f ./longhorn/longhorn-storage-backup-tmp.yaml 1>>"${dev_null}" || true
  kubectl apply -f ./longhorn/longhorn-storage-backup-tmp.yaml 1>>"${dev_null}"
  # create backup storage class with single replica

  rm -f ./longhorn/longhorn-storage-backup-one-replica-tmp.yaml &>>"${dev_null}"
  cp ./longhorn/longhorn-storage-backup-one-replica.yaml ./longhorn/longhorn-storage-backup-one-replica-tmp.yaml
  sed -i "s/REPLICA_PLACEHOLDER/${REPLICAS_BACKUP_ONE}/" ./longhorn/longhorn-storage-backup-one-replica-tmp.yaml
  sed -i "s/STALE_REPLICA_TIMEOUT_PLACEHOLDER/${STALE_REPLICA_TIMEOUT}/" ./longhorn/longhorn-storage-backup-one-replica-tmp.yaml
  # Delete older storageclass definitions so that new parameters are applied
  kubectl delete -f ./longhorn/longhorn-storage-backup-one-replica-tmp.yaml 1>>"${dev_null}" || true
  kubectl apply -f ./longhorn/longhorn-storage-backup-one-replica-tmp.yaml 1>>"${dev_null}"

  rm -f ./longhorn/longhorn-storage-tmp.yaml &>>"${dev_null}"
  cp ./longhorn/longhorn-storage.yaml ./longhorn/longhorn-storage-tmp.yaml
  sed -i "s/STALE_REPLICA_TIMEOUT_PLACEHOLDER/${STALE_REPLICA_TIMEOUT}/" ./longhorn/longhorn-storage-tmp.yaml
  # Delete older storageclass definitions so that new parameters are applied
  kubectl delete -f ./longhorn/longhorn-storage-tmp.yaml 1>>"${dev_null}" || true
  kubectl apply -f ./longhorn/longhorn-storage-tmp.yaml 1>>"${dev_null}"

  patch_existing_volumes_stale_timeout

  # create longhorn networkpolicy to block unauthenticated access
  if [[ "${USE_CILIUM}" -gt 0 ]]; then
    if kubectl -n longhorn-system get networkpolicy longhorn-networkpolicy &>>"${dev_null}";
    then
      # Delete the older networkpolicy
      kubectl -n longhorn-system delete networkpolicy longhorn-networkpolicy
    fi
  else
    kubectl apply -f ./longhorn/longhorn-networkpolicy.yaml &>>"${dev_null}"
  fi

  # create longhorn pdb for csi-attacher
  kubectl apply -f ./longhorn/longhorn-attacher-pdb.yaml &>>"${dev_null}"

  # create longhorn recurring snapshot if backup not enabled
  if ! kubectl -n longhorn-system get cronjob uipath-backup &>>"${dev_null}"; then
    rm -f ./longhorn/longhorn-recurring-snapshot-tmp.yaml &>>"${dev_null}"
    cp ./longhorn/longhorn-recurring-snapshot.yaml ./longhorn/longhorn-recurring-snapshot-tmp.yaml
    sed -i "s/SCHEDULE_PLACEHOLDER/${LH_RECURRING_SNAPSHOT_SCHEDULE}/g" ./longhorn/longhorn-recurring-snapshot-tmp.yaml
    kubectl apply -f ./longhorn/longhorn-recurring-snapshot-tmp.yaml &>>"${dev_null}"
  fi

  create_longhorn_stale_folder_cleanup_cronjob
  create_longhorn_loop_device_cleaner_daemonset


  # update longhorn csi-provisioner for 1.2.0 release https://github.com/longhorn/longhorn/issues/2964
  if [[ "${LONGHORN_VERSION}" == "1.2.0" ]]; then
    update_csi_provisioner
  fi
  info "Longhorn Installation Success"
}

function update_csi_provisioner() {
  local try=0
  local maxtry=10
  local csi_provisioner_exist=0

  while ((try < maxtry )); do
    result=$(kubectl -n longhorn-system get deploy csi-provisioner) || true
    if [[ -n "${result}" ]]; then
      csi_provisioner_exist=1
      break;
    fi

    try=$((try+1))
    info "Waiting for csi-provisioner deployment to come up. ${try}/${maxtry}..."; sleep 5
  done

  if (( csi_provisioner_exist == 0 ));  then
    error_with_code "11008"
  fi

  local container_data
  container_data=$(kubectl -n longhorn-system get deploy csi-provisioner -o json| jq -c ".spec.template.spec.containers[]| select( .name == \"csi-provisioner\").args[.args| length] |= . + \"--default-fstype=ext4\"")
  kubectl -n longhorn-system patch deployment csi-provisioner --patch "{\"spec\": {\"template\": {\"spec\": {\"containers\": [${container_data}]}}}}"
}

function test_longhorn(){
  kubectl delete -n longhorn-system -f ./longhorn/tests/pod_with_pvc.yaml --ignore-not-found=true &>>"${dev_null}"
  kubectl create -n longhorn-system -f ./longhorn/tests/pod_with_pvc.yaml &>>"${dev_null}"
  kubectl wait -n longhorn-system --for=condition=Ready --timeout=10m pods volume-test &>>"${dev_null}" || true

  POD=$(kubectl get pods -n longhorn-system volume-test \
    | grep volume-test \
    | grep Running \
    | grep "1/1")

  BOUND=$(kubectl get pvc -n longhorn-system -A \
    | grep longhorn-volv-pvc \
    | grep Bound)

  kubectl delete -n longhorn-system -f ./longhorn/tests/pod_with_pvc.yaml &>>"${dev_null}"

  [[ -n ${POD} ]] || {
    error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_STORAGE_SETUP_MESSAGE}";
    error_with_code "11009";
  }

  [[ -n ${BOUND} ]] || {
    error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_STORAGE_SETUP_MESSAGE}";
    error_with_code "11010";
  }
}

function install_docker_registry() {
  local helm_status
  info "Installing Docker registry"

  local tls_secret_name="docker-registry-tls"
  local namespace="docker-registry"
  local uipath_namespace="uipath"
  local auth_secret_name="docker-registry-credentials"
  declare -a DOCKER_REGISTRY_ARGS
  local current_secret

  # if INSTALL_TYPE is offline, then the registry image is already present on the machine
  # if INSTALL_TYPE is online, then images need to be pulled from DOCKER_REGISTRY_URL
  if [[ ${INSTALL_TYPE} == "online" ]]; then
    DOCKER_REGISTRY_ARGS+=(--set image.repository="${DOCKER_REGISTRY_URL}"/registry)
  fi

  kubectl get namespace ${namespace} \
    || kubectl create namespace ${namespace} &>>"${dev_null}" \
    || {  error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_STORAGE_SETUP_MESSAGE}";
          error_with_code "10003" "${namespace}";
        }

  secret=$(kubectl get secret "${tls_secret_name}" \
    --namespace "${namespace}" \
    --ignore-not-found \
    --no-headers)

  registry_dns_found="true"
  if [[ -n "${secret}" && "${INSTALL_TYPE}" == "offline" && "${ACTIVE_ACTIVE}" == "true" ]]; then
    secret=$(kubectl get secret "${tls_secret_name}" \
              --namespace "${namespace}" \
              --ignore-not-found \
              --no-headers -o jsonpath="{.data.tls\.crt}" | base64 -d)
    if [[ -n "${secret}" ]]; then
      tmp_dir=$(mktemp -d certificates.XXXXXXX)
      echo "${secret}" > "${tmp_dir}/tls.crt"
      tmp_dir_absolute_path=$(readlink -f "${tmp_dir}")
      response=$(validate_certificate_san "${tmp_dir_absolute_path}/tls.crt" "*.${ALTERNATIVE_FQDN}")
      if [[ -n ${response} ]]; then
        registry_dns_found="false"
        info "Docker registry SAN not configured in the cert"
      fi
      rm -rf "${tmp_dir}"
    fi
  fi

  if [[ "${secret}" == "" ]] || [[ "${UPDATE_FQDN}" == "true" ]] || [[ "${registry_dns_found}" == "false" ]];
  then
    # only update the certificate if it is not configured or updating fqdn
    info "configuring certificate in docker registry"

    #shellcheck disable=SC2129
    kubectl delete secret "${tls_secret_name}" \
    --namespace "${namespace}" --ignore-not-found &>>"${dev_null}"

    if [[ "${ACTIVE_ACTIVE}" == "true" ]]; then
      kubectl create secret generic ${tls_secret_name} \
        --namespace ${namespace} \
        --from-file=tls.crt="${ALT_CERT_TLS_FILE_PATH}" \
        --from-file=tls.key="${ALT_CERT_KEY_FILE_PATH}" &>>"${dev_null}"
    else
      kubectl create secret generic ${tls_secret_name} \
        --namespace ${namespace} \
        --from-file=tls.crt="${CERT_TLS_FILE_PATH}" \
        --from-file=tls.key="${CERT_KEY_FILE_PATH}" &>>"${dev_null}"
    fi
    kubectl label secret ${tls_secret_name} monitor-cert="true" \
      --namespace ${namespace} --overwrite=true &>>"${dev_null}"
  fi

# create docker registry secret if it doesn't exist
# if the secret is already present, it is not overwritten, but read
  current_secret=$(kubectl get secret ${auth_secret_name} -n ${uipath_namespace} --ignore-not-found)
  if [[ -z "${current_secret}" ]]; then
    info "Local docker registry secrets not found, creating..."
    DOCKER_REGISTRY_LOCAL_HTPASS="$(htpasswd -nBb "${DOCKER_REGISTRY_LOCAL_USERNAME}" "${DOCKER_REGISTRY_LOCAL_PASSWORD}")"
    kubectl create secret generic "${auth_secret_name}" \
        -n "${uipath_namespace}" \
        --from-literal=username="${DOCKER_REGISTRY_LOCAL_USERNAME}" \
        --from-literal=password="${DOCKER_REGISTRY_LOCAL_PASSWORD}" \
        --from-literal=htpass="${DOCKER_REGISTRY_LOCAL_HTPASS}" \
        --dry-run=client -o yaml | kubectl apply -f - &>>"${dev_null}"
  else
    info "Local Docker registry secrets already present, loading..."
    DOCKER_REGISTRY_LOCAL_USERNAME=$(kubectl get secret "${auth_secret_name}" -n "${uipath_namespace}" -o jsonpath="{.data.username}" | base64 -d)
    DOCKER_REGISTRY_LOCAL_PASSWORD=$(kubectl get secret "${auth_secret_name}" -n "${uipath_namespace}" -o jsonpath="{.data.password}" | base64 -d)
    DOCKER_REGISTRY_LOCAL_HTPASS=$(kubectl get secret "${auth_secret_name}" -n "${uipath_namespace}" -o jsonpath="{.data.htpass}" | base64 -d)
  fi

  DOCKER_REGISTRY_ARGS+=(--set tlsSecretName="${tls_secret_name}")
  DOCKER_REGISTRY_ARGS+=(--set secrets.htpasswd="${DOCKER_REGISTRY_LOCAL_HTPASS}")
  DOCKER_REGISTRY_ARGS+=(--set persistence.enabled="true")
  DOCKER_REGISTRY_ARGS+=(--set persistence.size="200Gi")
  DOCKER_REGISTRY_ARGS+=(--set service.nodePort="${DOCKER_REGISTRY_NODEPORT}")
  DOCKER_REGISTRY_ARGS+=(--set priorityClassName="system-cluster-critical")
  # enable deleting images
  DOCKER_REGISTRY_ARGS+=(--set configData.storage.delete.enabled="true")
  if [[ -n "${STORAGE_CLASS}" && "${STORAGE_CLASS}" != " " && "${STORAGE_CLASS}" != "null" ]]; then
    DOCKER_REGISTRY_ARGS+=(--set persistence.storageClass="${STORAGE_CLASS}")
  fi

  info "Installing Docker registry with arguments ${DOCKER_REGISTRY_ARGS[*]}"

  # fix for SERFAB-6580 0.5% the helm upgrade is stuck if SIGKILL or Ctrl+C
  # this check will rollback which is pretty safe since docker registry has never had a different version
  helm_status=$(helm status docker-registry --namespace docker-registry | grep "pending-upgrade") || true
  if [[ -n "${helm_status}"  ]]; then
    helm rollback docker-registry --namespace docker-registry || true
  fi

  helm upgrade --install --atomic docker-registry "${TMP_HELM_CHARTS_DIR}"/docker-registry/ \
    --wait --timeout 5m \
    --namespace docker-registry \
    --create-namespace \
    "${DOCKER_REGISTRY_ARGS[@]}" &>>"${dev_null}" || {
      error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_STORAGE_SETUP_MESSAGE}";
      error_with_code "11011";
    }
}

function install_istio() {
  info "Installing Istio"

  local namespace="istio-system"
  local ISTIOD_REPLICA_MIN_COUNT
  local ISTIOD_REPLICA_MAX_COUNT
  local PULL_SECRET_NAME="uipathpullsecret"
  local ENVOY_LUA_LIB_CM="envoy-lua-lib"

  kubectl get ns ${namespace} &>>"${dev_null}" || kubectl create ns ${namespace} &>>"${dev_null}"

  declare -a ISTIO_CHART_ARGS

  if [[ ${INSTALL_TYPE} == "online" ]]; then
    ISTIO_CHART_ARGS+=(--set global.cattle.systemDefaultRegistry="${DOCKER_REGISTRY_URL}")
  fi

  if [ -n "${DOCKER_REGISTRY_PULL_SECRET}" ]
  then
    sed -i "s;NAMESPACE_PLACEHOLDER;${namespace};"  ./rancher-istio/image-pull-secret.yaml
    sed -i "s;IMAGE_PULL_SECRET_PLACEHOLDER;${PULL_SECRET_NAME};" ./rancher-istio/image-pull-secret.yaml
    sed -i "s;IMAGE_PULL_SECRET_VALUE_PLACEHOLDER;${DOCKER_REGISTRY_PULL_SECRET};" ./rancher-istio/image-pull-secret.yaml
    kubectl apply -f ./rancher-istio/image-pull-secret.yaml &>>"${dev_null}"
    sed -i "s;LIST_IMAGE_PULL_SECRET_PLACEHOLDER;\[\"${PULL_SECRET_NAME}\"\];" ./rancher-istio/values-overlay.yaml
  else
    sed -i "s;LIST_IMAGE_PULL_SECRET_PLACEHOLDER;\[\];" ./rancher-istio/values-overlay.yaml
  fi

  if [[ "${PROFILE}" == "ha" ]]
  then
    ISTIOD_REQUESTS_CPU=500m
    ISTIOD_REQUESTS_MEMORY=2Gi
    ISTIOD_REPLICA_MIN_COUNT=2
    ISTIOD_REPLICA_MAX_COUNT=3
  else
    ISTIOD_REQUESTS_CPU=100m
    ISTIOD_REQUESTS_MEMORY=500Mi
    ISTIOD_REPLICA_MIN_COUNT=1
    ISTIOD_REPLICA_MAX_COUNT=1
  fi

  sed -i "s;NAMESPACE_PLACEHOLDER;${namespace};" ./rancher-istio/values-overlay.yaml
  sed -i "s;NAMESPACE_PLACEHOLDER;${namespace};" ./rancher-istio/values-overlay.yaml
  sed -i "s;ISTIOD_REQUESTS_CPU_PLACEHOLDER;${ISTIOD_REQUESTS_CPU};" ./rancher-istio/values-overlay.yaml
  sed -i "s;ISTIOD_REQUESTS_MEMORY_PLACEHOLDER;${ISTIOD_REQUESTS_MEMORY};" ./rancher-istio/values-overlay.yaml
  sed -i "s;ISTIOD_REPLICA_MIN_COUNT_PLACEHOLDER;${ISTIOD_REPLICA_MIN_COUNT};" ./rancher-istio/values-overlay.yaml
  sed -i "s;ISTIOD_REPLICA_MAX_COUNT_PLACEHOLDER;${ISTIOD_REPLICA_MAX_COUNT};" ./rancher-istio/values-overlay.yaml
  sed -i "s;ENVOY_LUA_LIB_CM_PLACEHOLDER;${ENVOY_LUA_LIB_CM};" ./rancher-istio/values-overlay.yaml

  kubectl -n ${namespace} create configmap "${ENVOY_LUA_LIB_CM}" \
    --from-file=./rancher-istio/lua-lib/src/  --dry-run=client -o yaml \
    | kubectl -n $namespace apply -f - &>>"${dev_null}"

  ## Delete istiod deployment to allow subsequent runs to pass
  ## Please follow this ticket for more update https://uipath.atlassian.net/browse/SERFAB-2071
  if kubectl --namespace $namespace get deployment/istiod
  then
    kubectl --namespace $namespace get deployment/istiod -o yaml > /tmp/istiod-deployment.yaml
    kubectl --namespace $namespace delete --wait=true deployment/istiod
  fi
  helm upgrade --install --atomic -f ./rancher-istio/values-overlay.yaml istio "${TMP_HELM_CHARTS_DIR}"/rancher-istio/ \
    --wait --timeout 10m \
    --namespace $namespace \
    --create-namespace \
    "${ISTIO_CHART_ARGS[@]}" &>>"${dev_null}" || {
      test -f /tmp/istiod-deployment.yaml && kubectl apply -f /tmp/istiod-deployment.yaml;
      error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_ROUTING_SETUP_MESSAGE}";
      error_with_code "11012";
    }

  info "Finished installing Istio"
}

function create_uipath_infra_namespace() {
  info "Creating uipath-infra namespace"
  local uipath_infra_namespace="uipath-infra"
  kubectl get namespace "${uipath_infra_namespace}" &>>"${dev_null}" \
  || kubectl create namespace "${uipath_infra_namespace}" &>>"${dev_null}" \
  || {
    error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_ROUTING_SETUP_MESSAGE}";
    error_with_code "10003" "${uipath_infra_namespace}";
  }
}

function create_uipath_namespace() {
  info "Creating uipath namespace"
  local uipath_namespace="uipath"
  kubectl get namespace "${uipath_namespace}" &>>"${dev_null}" \
  || kubectl create namespace "${uipath_namespace}" &>>"${dev_null}" \
  || {
    error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_ROUTING_SETUP_MESSAGE}";
    error_with_code "10003" "${uipath_namespace}";
  }
}

function create_istio_namespace() {
  info "Creating Istio namespace"
  local istio_namespace="istio-system"
  kubectl get namespace "${istio_namespace}" &>>"${dev_null}" \
  || kubectl create namespace "${istio_namespace}" &>>"${dev_null}" \
  || {
    error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_ROUTING_SETUP_MESSAGE}";
    error_with_code "10003" "${istio_namespace}";
  }
}

function configure_istio() {
  info "Configuring Istio"

  local infra_namespace="uipath-infra"
  local istio_namespace="istio-system"
  local istio_tls_cert_name="istio-ingressgateway-certs"
  local istio_tls_alt_cert_name="istio-ingressgateway-alt-certs"
  local ca_cert_name="ca.crt"
  local tls_cert_name="tls.crt"
  local tls_key_name="tls.key"
  local label='config-discovery=yes'

  secret=$(kubectl get secret "${istio_tls_cert_name}" \
    --namespace "${istio_namespace}" \
    --ignore-not-found \
    --no-headers)

  if [[ "${secret}" == "" ]] || [[ "${UPDATE_FQDN}" == "true" ]];
  then
    # only update certificate if not already configured or updating fqdn
    info "configuring certificate in istio / cluster"
    tmp_dir=$(mktemp -d certificates.XXXXXXX)
    tmp_file="ca_tls_merged.crt"
    add_empty_line_in_file_if_not_exists "${CERT_TLS_FILE_PATH}"

    cat "${CERT_TLS_FILE_PATH}" "${CERT_CA_FILE_PATH}" >> "${tmp_dir}"/"${tmp_file}"
    merge_ca_and_tls_path="${tmp_dir}"/"${tmp_file}"

    kubectl create secret generic "${istio_tls_cert_name}" \
      --from-file="${ca_cert_name}"="${CERT_CA_FILE_PATH}" \
      --from-file="${tls_cert_name}"="${merge_ca_and_tls_path}" \
      --from-file="${tls_key_name}"="${CERT_KEY_FILE_PATH}" \
      --namespace "${istio_namespace}" \
      --dry-run=client -o yaml | kubectl apply -f - &>>"${dev_null}"
    
    rm -rf "${tmp_dir}"
    
    kubectl label secret ${istio_tls_cert_name} "${label}" \
      --namespace ${istio_namespace} --overwrite=true &>>"${dev_null}"

    kubectl label secret ${istio_tls_cert_name} monitor-cert="true" \
      --namespace ${istio_namespace} --overwrite=true &>>"${dev_null}"
  fi

  if [[ -n "${ALTERNATIVE_FQDN}" ]]; then
    tmp_dir=$(mktemp -d certificates.XXXXXXX)
    alt_secret=$(kubectl get secret "${istio_tls_alt_cert_name}" \
      --namespace "${istio_namespace}" \
      --ignore-not-found \
      --no-headers -o jsonpath="{.data.tls\.crt}")
    dns_found="false"
    if [[ -n "${alt_secret}" ]]; then
      echo "${alt_secret}" | base64 -d > "${tmp_dir}/tls.crt"
      csplit -s -z -f "${tmp_dir}/pem-" "${tmp_dir}/tls.crt" '/-----BEGIN CERTIFICATE-----/' '{*}'
      tmp_dir_absolute_path=$(readlink -f "${tmp_dir}")
      for i in "${tmp_dir_absolute_path}"/pem-*; do
        info "Checking cert for SAN: ${i}"
        response=$(validate_certificate_san "${i}" "*.${ALTERNATIVE_FQDN}")
        [[ -z ${response} ]] && {
          dns_found="true"
          info "SAN configured in the cert"
          break
        }
      done
      rm -rf "${tmp_dir_absolute_path}"
    fi

    if [[ "${DO_UPGRADE}" == "0" ]] && [[ -z "${alt_secret}" || ("${UPDATE_FQDN}" == "true" || "${dns_found}" == "false" ) ]]; then
      # only update certificate if not already configured or updating fqdn
      info "configuring alt_fqdn certificate in istio / cluster"
      tmp_dir=$(mktemp -d certificates.XXXXXXX)
      tmp_file="ca_tls_merged.crt"
      cat "${ALT_CERT_TLS_FILE_PATH}" "${ALT_CERT_CA_FILE_PATH}" >> "${tmp_dir}"/"${tmp_file}"
      merge_ca_and_tls_path="${tmp_dir}"/"${tmp_file}"

      kubectl create secret generic "${istio_tls_alt_cert_name}" \
        --from-file="${ca_cert_name}"="${ALT_CERT_CA_FILE_PATH}" \
        --from-file="${tls_cert_name}"="${merge_ca_and_tls_path}" \
        --from-file="${tls_key_name}"="${ALT_CERT_KEY_FILE_PATH}" \
        --namespace "${istio_namespace}" \
        --dry-run=client -o yaml | kubectl apply -f - &>>"${dev_null}"

      rm -rf "${tmp_dir}"

      kubectl label secret ${istio_tls_alt_cert_name} "${label}" \
        --namespace ${istio_namespace} --overwrite=true &>>"${dev_null}"

      kubectl label secret ${istio_tls_alt_cert_name} monitor-cert="true" \
        --namespace ${istio_namespace} --overwrite=true &>>"${dev_null}"
    fi
  fi

  kubectl get namespace "${infra_namespace}" &>>"${dev_null}" \
    || kubectl create namespace "${infra_namespace}" &>>"${dev_null}" \
    || {
      error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_ROUTING_SETUP_MESSAGE}";
      error_with_code "10003" "${infra_namespace}";
    }

  declare -a ISTIO_CONFIGURE_CHART_ARGS

  ISTIO_CONFIGURE_CHART_ARGS+=(--set fqdn="${CLUSTER_FQDN}")
  # if CLUSTER_FQDN and ALTERNATIVE_FQDN are the same, don't pass ALTERNATIVE_FQDN to prevent duplicate hostnames.
  [[ "${CLUSTER_FQDN}" != "${ALTERNATIVE_FQDN}" ]] && ISTIO_CONFIGURE_CHART_ARGS+=(--set alt_fqdn="${ALTERNATIVE_FQDN}") || ISTIO_CONFIGURE_CHART_ARGS+=(--set alt_fqdn="")
  ISTIO_CONFIGURE_CHART_ARGS+=(--set ingresses.insightslooker.fqdn="${LOOKER_SUBDOMAIN}")
  ISTIO_CONFIGURE_CHART_ARGS+=(--set ingresses.cephobjectstore.fqdn="${OBJECTSTORE_SUBDOMAIN}")
  ISTIO_CONFIGURE_CHART_ARGS+=(--set ingresses.rancherserver.fqdn="${RANCHER_SUBDOMAIN}")
  ISTIO_CONFIGURE_CHART_ARGS+=(--set ingresses.argocdserver.fqdn="${ARGO_SUBDOMAIN}")
  ISTIO_CONFIGURE_CHART_ARGS+=(--set ingresses.dockerregistry.fqdn="${DOCKER_REGISTRY_SUBDOMAIN}")

  if [[ ${INSTALL_TYPE} == "online" ]]; then
    ISTIO_CONFIGURE_CHART_ARGS+=(--set uipath.registry="${DOCKER_REGISTRY_URL}")
  else
    ISTIO_CONFIGURE_CHART_ARGS+=(--set uipath.registry="")
  fi

  if [[ ${INSTALLSTANDALONEAICENTER} == "1" ]]; then
      ISTIO_CONFIGURE_CHART_ARGS+=(--set orchestratorTrafficPolicies.enabled="false" )
      ISTIO_CONFIGURE_CHART_ARGS+=(--set firstPartyTrafficPolicies.enabled="false" )
      ISTIO_CONFIGURE_CHART_ARGS+=(--set identityTrafficPolicies.enabled="false" )
  fi

  if [[ ${ACTIVE_ACTIVE} == "true" ]]; then
      if [[ ${PRIMARY_CLUSTER} == "false" ]]; then
        [[ -z "${PRIMARY_CLUSTER_FQDN}" ]] && error_with_code "15013"
        ISTIO_CONFIGURE_CHART_ARGS+=(--set requestForwarding.target="${PRIMARY_CLUSTER_FQDN}" )
      fi
  fi

  [ -z "${DOCKER_REGISTRY_PULL_SECRET}" ] || ISTIO_CONFIGURE_CHART_ARGS+=(--set global.imagePullSecret.value="${DOCKER_REGISTRY_PULL_SECRET}")

  helm upgrade --install --atomic istio-configure "${TMP_HELM_CHARTS_DIR}"/istio-configure/ \
    --wait --timeout 10m \
    --namespace $infra_namespace \
    --create-namespace \
    "${ISTIO_CONFIGURE_CHART_ARGS[@]}" &>>"${dev_null}" || {
      error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_ROUTING_SETUP_MESSAGE}";
      error_with_code "11013";
    }

    info "Finished Configuring Istio"
}

function test_istio_installation() {
  local fqdn
  [[ ${ALTERNATIVE_FQDN} == https://* ]] && fqdn="${ALTERNATIVE_FQDN}" || fqdn="https://${ALTERNATIVE_FQDN}"
  info "Going to test istio on endpoint ${fqdn}"

  retry=0
  until [ "$retry" -ge 30 ]
  do
    response=$(curl -i -k -sS -m 10 "${fqdn}") || warn "Istio test failed, Retry no ${retry}...."
    grep -q "istio" <<< "${response}" && info "Istio test passed" && return
    retry=$((retry+1))
    sleep 10
  done
  error_with_code "11023" "${retry}"
}

function validate_subdomain_resolution() {
  local domain
  local domain_ip
  local pod_domain_ip
  local istio_namespace="istio-system"
  local error_messages=()
  local warn_messages=()

  while IFS=$'\n' read -r subdomain_resolve_map_iter
  do
    domain=$(echo "${subdomain_resolve_map_iter}" | cut -d'=' -f1)
    domain_ip=$(echo "${subdomain_resolve_map_iter}" | cut -d'=' -f2)
    pod_domain_ip=$(kubectl -n "${istio_namespace}" exec ds/istio-ingressgateway -- getent ahosts  "${domain}" 2>>"${dev_null}" | awk '{print $1}' | sort | uniq | tr '\n' ',' | sed 's/,$//g')
    if [[ -z "${pod_domain_ip}" ]]
    then
      error_messages+=("No IP Address found for ${domain} from container. Resolved to ${domain_ip} from host")
    elif [[ "$pod_domain_ip" != "$domain_ip" ]]
    then
      warn_messages+=("Resolved ip(s) for ${domain} not same when tried from host(${domain_ip}) and container(${pod_domain_ip})")
    fi
  done < /tmp/subdomain_resolve_map.txt
  if [[ ${#warn_messages[@]} != 0 ]] ; then
    warn_output "Discrepancy found in DNS config from container" "${warn_messages[@]}"
  fi
  if [[ ${#error_messages[@]} != 0 ]] ; then
    failure_output "DNS Validation Failed from container" "${error_messages[@]}"
  fi
}


function store_identity_certificate() {
  local namespace="uipath"
  local secret_name="identity-token-signing-certificate"
  local cert_name_1="IdentityServer1.pfx"
  local pass_1="password1"
  local cert_name_2="IdentityServer2.pfx"
  local pass_2="password2"

  # create uipath-infra namespace and store the certificate in secret
  kubectl get namespace "${namespace}" \
    || kubectl create namespace "${namespace}" \
    || {  error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_ROUTING_SETUP_MESSAGE}";
          error_with_code "10003" "${namespace}";
       }

  kubectl get secret "${secret_name}" --namespace "${namespace}" \
    || kubectl create secret generic "${secret_name}" \
        --from-file="${cert_name_1}"="${TOKEN_SIGNING_CERT}" \
        --from-literal="${pass_1}"="${TOKEN_SIGNING_PASS}" \
        --from-file="${cert_name_2}"="${TOKEN_SIGNING_CERT}" \
        --from-literal="${pass_2}"="${TOKEN_SIGNING_PASS}" \
        --namespace "${namespace}" \
        --dry-run=client -o yaml | kubectl apply -f - \
    || {  error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_ROUTING_SETUP_MESSAGE}";
          error_with_code "11024";
       }

  info "Token signing certificate is stored"
}

function store_additional_ca_certs() {
  local namespace="uipath"
  local secret_name="additional-ca-certificates"
  local cert_file="additionalca.crt"

  if [[ -f "${ADDITIONAL_CA_CERTS}" ]];
  then
    # create uipath namespace and store the certificate in secret
    kubectl get namespace "${namespace}" \
      || kubectl create namespace "${namespace}" \
      || {  error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_ROUTING_SETUP_MESSAGE}";
            error_with_code "10003" "${namespace}";
        }

    kubectl get secret "${secret_name}" --namespace "${namespace}" \
      || kubectl create secret generic "${secret_name}" \
          --from-file="${cert_file}"="${ADDITIONAL_CA_CERTS}" \
          --namespace "${namespace}" \
          --dry-run=client -o yaml | kubectl apply -f - \
      || {  error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_ROUTING_SETUP_MESSAGE}";
            error_with_code "11025";
        }

    info "Additional trusted CA certificates are stored"
  fi
}

function test_server_api(){
  local try=0
  local maxtry=120

  while [[ ! -f "${KUBECONFIG}" ]] &&  (( try != maxtry )); do
          try=$((try+1))
          info "Waiting for RKE2 kubeconfig... ${try}/${maxtry}" &&  sleep 10;
  done

  [[ -f "${KUBECONFIG}" ]] || {
    error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_K8S_INSTALL_MESSAGE}";
    error_with_code "11014"
  }

  try=0
  maxtry=60
  while [[ $(kubectl get --raw='/readyz' 2>>"${dev_null}") != "ok" ]] && (( try != maxtry )); do
    try=$((try+1))
    info "waiting for kubernetes api server to be ready...${try}/${maxtry}" &&  sleep 10;
  done

  local hostname_lower
  hostname_lower="$(hostname | awk '{print tolower($0)}')"
  try=0
  while [[ $(kubectl get node "${hostname_lower}" -o 'jsonpath={..status.conditions[?(@.type=="Ready")].status}' 2>>"${dev_null}") =~ "False" ]] && (( try != maxtry )); do
    try=$((try+1))
    info "waiting for server to be ready...${try}/${maxtry}" &&  sleep 10;
  done

  kubectl cluster-info &>>"${dev_null}" || {
    error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_K8S_INSTALL_MESSAGE}";
    error_with_code "10004";
  }
}

function test_agent_api(){
  local try=0
  local maxtry=120
  local kubeconfig="/var/lib/rancher/rke2/agent/kubelet.kubeconfig"

  while [[ ! -f "${kubeconfig}" ]] &&  (( try != maxtry )); do
          try=$((try+1))
          info "Waiting for RKE2 kubeconfig... ${try}/${maxtry}" &&  sleep 10;
  done

  [[ -f "${kubeconfig}" ]] || error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_K8S_INSTALL_MESSAGE}"
  [[ -f "${kubeconfig}" ]] || error_with_code "11014"

  try=0
  maxtry=60

  while [[ $(kubectl --kubeconfig "${kubeconfig}" get --raw='/readyz' 2>>"${dev_null}") != "ok" ]] && (( try != maxtry )); do
    try=$((try+1))
    info "waiting for kubernetes api server to be ready...${try}/${maxtry}" &&  sleep 10;
  done

  local hostname_lower
  hostname_lower="$(hostname | awk '{print tolower($0)}')"

  try=0
  while [[ $(kubectl --kubeconfig "${kubeconfig}" get node "${hostname_lower}" -o 'jsonpath={..status.conditions[?(@.type=="Ready")].status}' 2>>"${dev_null}") =~ "False" ]] && (( try != maxtry )); do
    try=$((try+1))
    info "waiting for agent to be ready...${try}/${maxtry}" &&  sleep 10;
  done

  kubectl --kubeconfig "${kubeconfig}" cluster-info &>>"${dev_null}" || error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_K8S_INSTALL_MESSAGE}"
  kubectl --kubeconfig "${kubeconfig}" cluster-info &>>"${dev_null}" || error_with_code "10004"
}

function set_output() {
  cat > "${OUTFILE}" <<EOF
{
  "docker_registry_user": "${DOCKER_REGISTRY_LOCAL_USERNAME}",
  "docker_registry_password": "${DOCKER_REGISTRY_LOCAL_PASSWORD}"
}
EOF
}


function create_priority_classes(){
    kubectl apply -f "${SCRIPT_DIR}/k8s_manifests/priority_class.yaml"
}

function send_input_telemetry()  {
  if [[ "${TELEMETRY_OPT_OUT:-false}" == "true" ]]; then
    info "Telemetry is off"
    return;
  fi

  local now
  now=$(date -u +%Y-%m-%dT%H:%M:%S.%3NZ)

  ensure_machine_hash

  function data() {
    local telemetry_content

    cat <<EOF
{
    "name": "AppEvents",
    "time": "$now",
    "iKey": "$TELEMETRY_INSTRUMENTATION_KEY",
    "tags": {
        "ai.cloud.role": "Installer",
        "ai.internal.sdkVersion": "REST"
    },
    "data": {
        "baseType": "EventData",
        "baseData": {
            "ver": 2,
            "name": "Infra.Input",
            "properties": {
              "ApplicationName": "UiPath.ServiceFabric.Installer",
              "ProductVersion": "${VERSION}",
              "InstallId": "${INSTALL_ID:-none}",
              "NodeRole" : "${NODE_ROLE}",
              "NodeType" : "${NODE_TYPE}",
              "EtcdRestore" : "${ETCD_RESTORE}",
              "InstallType" : "${INSTALL_TYPE}",
              "ClusterToken" : "${RKE_TOKEN}",
              "CloudTemplateVendor" : "${CLOUD_TEMPLATE_VENDOR}",
              "CloudTemplateSource" : "${CLOUD_TEMPLATE_SOURCE}",
              "MachineHash" : "${MACHINE_HASH}",
              "LonghornVersion" : "${LONGHORN_VERSION}",
              "LonghornReplicaCount" : "${REPLICAS}",
              "LonghornHAReplicaCount" : "${HA_REPLICAS}",
              "BackupInterval" : "${BACKUP_INTERVAL}",
              "StorageClass" : "${STORAGE_CLASS}",
              "SystemReservedCPU" : "${SYSTEM_RESERVED_CPU}",
              "SystemReservedMemory" : "${SYSTEM_RESERVED_MEMORY}",
              "SystemReservedEphemeralStorage" : "${SYSTEM_RESERVED_EPHEMERAL_STORAGE}",
              "KubeReservedCPU" : "${KUBE_RESERVED_CPU}",
              "KubeReservedMemory" : "${KUBE_RESERVED_MEMORY}",
              "KubeReservedEphemeralStorage" : "${KUBE_RESERVED_EPHEMERAL_STORAGE}",
              "EvictionHardMemoryAvailable" : "${EVICTION_HARD_MEMORY_AVAILABLE}",
              "EvictionHardNodeFSAvailable" : "${EVICTION_HARD_NODEFS_AVAILABLE}",
              "ProxyEnabled" : "${PROXY_ENABLED}"
            }
        }
    }
}
EOF
  }

  telemetry_content="$(data)"
  send_telemetry "${telemetry_content}"
}

function send_hardware_telemetry() {
  local telemetry_content

  if [[ "${TELEMETRY_OPT_OUT:-false}" == "true" ]]; then
    info "Telemetry is off"
    return;
  fi

  local cpu_dimension
  local memory_total_dim
  local memory_available_dim
  local gpu_vga_dim
  local gpu_3d_dim
  local gpu_display_dim
  local disk_dim
  local virtualization_dim
  local kernel_dim
  local os_dim

  # CPU
  cpu_dimension=$(lscpu | grep -v Flags | tr '\n' ',')

  # Memory
  memory_total_dim=$(free -m | grep Mem: | awk '{print $2}')
  memory_available_dim=$(free -m | grep Mem: | awk '{print $7}')

  # GPU
  gpu_vga_dim=$(lspci | grep ' VGA ' | tr '\n' ',')
  gpu_3d_dim=$(lspci | grep ' 3D ' | tr '\n' ',')
  gpu_display_dim=$(lspci | grep ' Display ' | tr '\n' ',')

  # Disk
  disk_dim=$(lsblk -ido KNAME,TYPE,SIZE,MODEL | tr '\n' ',')

  # Virtualization
  virtualization_dim=$(hostnamectl | grep Virt | cut -d: -f2 | xargs)

  # Kernel + Architecture
  kernel_dim=$(uname -mrs)

  # Operating System
  os_dim=$(hostnamectl | grep 'Operating System' | cut -d: -f2 | xargs)

  # Azure
  csp_dim=$(curl -s -m 3 -H Metadata:true --noproxy "*" "http://169.254.169.254/metadata/instance?api-version=2021-01-01" | jq -r '.compute.azEnvironment' 2>>"${dev_null}") || true

  if [[ "${#csp_dim}" -lt 1 ]]; then
    # GCP
    csp_dim=$(curl -s -m 3 "http://169.254.169.254/computeMetadata/v1/instance/?recursive=true&alt=json" -H "Metadata-Flavor: Google" | jq -r '.cpuPlatform' 2>>"${dev_null}") || true

    if [[ "${#csp_dim}" -lt 1 ]]; then
      # AWS: IMDSv1
      csp_dim=$(curl -s -m 3 "http://169.254.169.254/latest/meta-data/ami-id" 2>>"${dev_null}") || true
      if [[ "${#csp_dim}" -lt 1 ]]; then
        # AWS: IMDSv2
        aws_csp_token=$(curl -s -m 3 -X PUT "http://169.254.169.254/latest/api/token" -H "X-aws-ec2-metadata-token-ttl-seconds: 21600" 2>>"${dev_null}") || true
        csp_dim=$(curl -s -m 3 -H "X-aws-ec2-metadata-token: $aws_csp_token" -v http://169.254.169.254/latest/meta-data/ 2>>"${dev_null}") || true
      fi

      if [[ "${#csp_dim}" -lt 1 ]]; then
        csp_dim="(unknown)"
      else
        csp_dim="AWS"
      fi
    else
      csp_dim="GCP"
    fi
  else
    csp_dim="Azure"
  fi

  local now
  now=$(date -u +%Y-%m-%dT%H:%M:%S.%3NZ)

  ensure_machine_hash

  function data() {
    cat <<EOF
{
    "name": "AppEvents",
    "time": "$now",
    "iKey": "$TELEMETRY_INSTRUMENTATION_KEY",
    "tags": {
        "ai.cloud.role": "Installer",
        "ai.internal.sdkVersion": "REST"
    },
    "data": {
        "baseType": "EventData",
        "baseData": {
            "ver": 2,
            "name": "Infra.HardwareInfo",
            "properties": {
                "ApplicationName": "UiPath.ServiceFabric.Installer",
                "InstallId": "${INSTALL_ID:-none}",
                "MachineHash" : "${MACHINE_HASH}",
                "ClusterToken" : "${RKE_TOKEN}",
                "CPU": "${cpu_dimension}",
                "Memory.Total": "${memory_total_dim}",
                "Memory.Available": "${memory_available_dim}",
                "GPU.VGA": "${gpu_vga_dim}",
                "GPU.3D": "${gpu_3d_dim}",
                "GPU.Display": "${gpu_display_dim}",
                "Disk": "${disk_dim}",
                "Virtualization": "${virtualization_dim}",
                "Kernel": "${kernel_dim}",
                "OS": "${os_dim}",
                "CloudServiceProvider": "${csp_dim:-unknown}"
            }
        }
    }
}
EOF
  }

  telemetry_content="$(data)"
  send_telemetry "${telemetry_content}"
}

function rke2_coredns_update() {
  local active_active="${1}"
  local private_fqdn="${2:-}"
  local tm_fqdn="${3:-}"

  local try=0
  local maxtry=20
  local wait=30
  local status="notready"

  local kube_system_namespace="kube-system"
  declare -a CUSTOM_CORE_DNS_CHART_ARGS

  # Delete rke2-coredns-pdb.yaml file from /var/lib/rancher/rke2/server/manifests/ as it will
  # be managed by helm going forward
  if [[ "${DO_UPGRADE}" == "1" ]]; then
    if [[ "${NODE_ROLE}" == "none" || "${NODE_ROLE}" == "server" ]]; then
      local rke2CoreDnsPdbConfigurationPath="/var/lib/rancher/rke2/server/manifests/rke2-coredns-pdb.yaml"
      if [[ -e "${rke2CoreDnsPdbConfigurationPath}" ]]; then
        info "Removing ${rke2CoreDnsPdbConfigurationPath} as it will be managed by custom HelmChartConfig"
        # annotate and label current rke2-coredns HelmChartConfig so it can be managed by helm going forward
        kubectl --namespace "$kube_system_namespace" annotate HelmChartConfig rke2-coredns meta.helm.sh/release-name=rke2-coredns-config --overwrite=true
        kubectl --namespace "$kube_system_namespace" annotate HelmChartConfig rke2-coredns meta.helm.sh/release-namespace="$kube_system_namespace" --overwrite=true
        kubectl --namespace "$kube_system_namespace" label HelmChartConfig rke2-coredns app.kubernetes.io/managed-by=Helm --overwrite=true
        rm -f "${rke2CoreDnsPdbConfigurationPath}"
        systemctl restart rke2-server.service || {
          error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_COREDNS_SETUP_MESSAGE}";
          error_with_code "11004";
        }
      fi
    fi
  fi

  if [[ $active_active == "true" ]]; then
    CUSTOM_CORE_DNS_CHART_ARGS+=(--set tm_fqdn="${tm_fqdn}")
    CUSTOM_CORE_DNS_CHART_ARGS+=(--set fqdn="${private_fqdn}")
    CUSTOM_CORE_DNS_CHART_ARGS+=(--set active_active.enabled="true")
  fi

  while [[ ${status} == "notready" ]] && (( try != maxtry )); do
    try=$((try+1))
    info "Installing rke2-coredns-config helm chart"
    helm upgrade --install --atomic rke2-coredns-config "${TMP_HELM_CHARTS_DIR}"/rke2-coredns-config/ \
      --wait --timeout 10m \
      --namespace "$kube_system_namespace" \
      "${CUSTOM_CORE_DNS_CHART_ARGS[@]}" &>>"${dev_null}" && status="ready"
    [[ ${status} != "ready" ]] && info "Retry ${try}..." && sleep ${wait}
  done

  if [[ "${status}" != "ready" ]]; then
    error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_COREDNS_SETUP_MESSAGE}";
    error_with_code "14023";
  else
    info "Finished Configuring custom RKE2 CoreDNS"
  fi
}

function store_cluster_configuration() {
  update_cluster_config "cluster_version" "$VERSION"
  update_cluster_config "cluster_fqdn" "$CLUSTER_FQDN"
  update_cluster_config "alternative_fqdn" "$ALTERNATIVE_FQDN"
  update_cluster_config "argocd_fqdn" "$ARGOCD_FQDN"
  update_cluster_config "cluster_port" "$CLUSTER_PORT"
  update_cluster_config "monitoring_fqdn" "$RANCHER_FQDN"
  update_cluster_config "objectstore_fqdn" "$CEPH_FQDN"
  if [[ ${INSTALL_TYPE} == "offline" ]]; then
    update_cluster_config "registry_fqdn" "$DOCKER_REGISTRY_FQDN"
    update_cluster_config "registry_nodeport" "$DOCKER_REGISTRY_NODEPORT"
    update_cluster_config "podman_folder" "$PODMAN_FOLDER"
  fi
  update_cluster_config "insights_fqdn" "$LOOKER_FQDN"
  update_cluster_config "fixed_rke_address" "$FIXED_RKE_ADDRESS"
  update_cluster_config "install_type" "$INSTALL_TYPE"
  update_cluster_config "rke_token" "$RKE_TOKEN"
  update_cluster_config "profile" "$PROFILE"

  [[ -n ${DOCKER_REGISTRY_USERNAME} ]] && update_cluster_config "registries.docker.username" "$DOCKER_REGISTRY_USERNAME"
  [[ -n ${DOCKER_REGISTRY_PASSWORD} ]] && update_cluster_config "registries.docker.password" "$DOCKER_REGISTRY_PASSWORD"
  #Update service installation type - standalone aicenter or all services
  local service_install_type
  service_install_type=$(get_cluster_config "SERVICE_INSTALL_TYPE")

  if [[ -z $service_install_type ]];
  then
    value="all"
    [[ ${INSTALLSTANDALONEAICENTER} == "1" ]] && value="aicenter"
    update_cluster_config "SERVICE_INSTALL_TYPE" "$value"
  elif [[ "${service_install_type}" == "all" ]] && [[ ${INSTALLSTANDALONEAICENTER} == "1" ]];
  then
    error_with_code "11022"
  elif [[ "${service_install_type}" == "aicenter" ]] && [[ ${INSTALLSTANDALONEAICENTER} == "0" ]];
  then
    error_with_code "11021"
  fi

  # hadr entries
  if [[ "${ACTIVE_ACTIVE:-false}" == "true" ]]; then
    update_cluster_config "active_active.enabled" "true"
    if [[ "${PRIMARY_CLUSTER:-false}" == "true" ]]; then
      update_cluster_config "active_active.primary" "true"
    else
      alternative_cluster_kubeconfig="$(get_mandatory_json_value ".active_active.primary_kube_config")"
      update_cluster_config "active_active.primary" "false"
      update_cluster_config ".active_active.primary_kube_config" "${alternative_cluster_kubeconfig}"
    fi
  fi
}

function add_node_taint_config() {
  local config_path=$1
  taints=""
  if [[ "${NODE_TYPE}" == "task-mining" ]]; then
    taints="task.mining/cpu=present:NoSchedule"
  elif [[ "${NODE_TYPE}" == "gpu" ]]; then
    taints="nvidia.com/gpu=present:NoSchedule"
  elif [[ "${NODE_TYPE}" == "serverless" ]]; then
    taints="serverless.robot=present:NoSchedule"
  fi

  if [[ "${taints}" != "" ]]; then
    cat >> "${config_path}" <<< "node-taint: ${taints}"
  fi
}

function process_primary_cluster_kcfg() {
  alternative_cluster_kubeconfig="$(get_mandatory_json_value ".active_active.primary_kube_config")"
  [[ -z "${alternative_cluster_kubeconfig}" ]] && error_with_code "15013"

  tmp_kubeconfig_file="kubeconfig.yaml"
  alternative_cluster_decoded_kubeconfig_path="${SCRIPT_DIR}/${tmp_kubeconfig_file}"

  echo "${alternative_cluster_kubeconfig}"  | tr -d '"' | base64 --decode > "${alternative_cluster_decoded_kubeconfig_path}" || \
    error_with_code "15012"

  cluster_fqdn="$(get_mandatory_json_value ".fqdn")"
  alt_fqdn="$(get_optional_json_value ".alternative_fqdn")"

  if [[ -z "${alt_fqdn}" ]]; then
    alt_fqdn="$(get_mandatory_json_value ".fqdn")"
  fi
  info "Alternative FQDN: ${alt_fqdn}"
  info "Cluster FQDN: ${cluster_fqdn}"

  kubectl -n uipath-infra create secret generic alternate-cluster-access --from-file=alternate-cluster-kubeconfig="${alternative_cluster_decoded_kubeconfig_path}" \
    --dry-run=client -o yaml | kubectl apply -f - || error_with_code "15009"

  rm -f "${alternative_cluster_decoded_kubeconfig_path}"
  return 0
}

function  setup_ceph_tls_primary() {
  info "configuring certificate for ceph ingress"

  # TODO: only generate certs if this is the primary cluster
  tmp_dir=$(mktemp -d ceph-certificates.XXXXXXX) || error "Failed to create temporary ceph cert directory"
  merged_ca_tls_file="ca_tls_merged.crt"
  ceph_ca_file_path="${tmp_dir}/ceph-ca.crt"
  ceph_tls_file_path="${tmp_dir}/ceph-tls.crt"
  ceph_key_file_path="${tmp_dir}/ceph-tls.key"
  local ca_cert_name="ca.crt"
  local tls_cert_name="tls.crt"
  local tls_key_name="tls.key"
  local istio_namespace="istio-system"

  create_self_issued_certificate "objectstore.${ALTERNATIVE_FQDN}" "${ceph_ca_file_path}" \
    "${ceph_tls_file_path}" "${ceph_key_file_path}" "36501" \
    "36500" "36400" 1 "${tmp_dir}" ""

  merge_ca_and_tls_path="${tmp_dir}/${merged_ca_tls_file}"
  add_empty_line_in_file_if_not_exists "${ceph_tls_file_path}"

  cat "${ceph_tls_file_path}" "${ceph_ca_file_path}" >> "${merge_ca_and_tls_path}"

  # create secret in istio namespace for ingress
  kubectl create secret generic "rook-ceph-replication-tls" \
    --from-file="${ca_cert_name}"="${ceph_ca_file_path}" \
    --from-file="${tls_cert_name}"="${merge_ca_and_tls_path}" \
    --from-file="${tls_key_name}"="${ceph_key_file_path}" \
    --namespace "${istio_namespace}" \
    --dry-run=client -o yaml | kubectl apply -f - &>>"${dev_null}"

  ca_key_name="ca_key"
  ca_key_pass_name="ca_key_pass"
  ca_key_file="${tmp_dir}/intermediate.ca.key"
  ca_key_pass_file=${tmp_dir}/intermediate.ca.key.pass
  kubectl create secret generic "rook-ceph-ingress-ca-key" \
  --from-file="${ca_key_name}"="${ca_key_file}" \
  --from-file="${ca_key_pass_name}"="${ca_key_pass_file}" \
  --namespace "${istio_namespace}" \
  --dry-run=client -o yaml | kubectl apply -f - &>>"${dev_null}"

  # create rook-ceph namespace
  kubectl get namespace "rook-ceph" \
    || kubectl create namespace "rook-ceph" \
    || error "Failed to create namespace rook-ceph"

  caundle_name="cabundle"
  # create caBundle in rook-ceph namespace
  kubectl create secret generic "rook-ceph-ingress-ca" \
    --from-file="${caundle_name}"="${ceph_ca_file_path}" \
    --namespace "rook-ceph" \
    --dry-run=client -o yaml | kubectl apply -f - &>>"${dev_null}"

  rm -rf "${tmp_dir}"
}

function setup_ceph_tls_secondary {
  tmp_dir=$(mktemp -d ceph-certificates.XXXXXXX) || error "Failed to create temporary ceph cert directory"
  merged_ca_tls_file="ca_tls_merged.crt"
  ceph_ca_file_path="${tmp_dir}/ceph-ca.crt"
  ceph_tls_file_path="${tmp_dir}/ceph-tls.crt"
  ceph_key_file_path="${tmp_dir}/ceph-tls.key"
  local istio_namespace="istio-system"
  local rook_ceph_namespace="rook-ceph"
  local ca_cert_name="ca.crt"
  local tls_cert_name="tls.crt"
  local tls_key_name="tls.key"
  tmp_kubeconfig_file="kubeconfig.yaml"
  primary_cluster_kcfg="${tmp_dir}/primary-cluster-kubeconfig.yaml"

  kubectl get secret -n uipath-infra alternate-cluster-access -o jsonpath='{.data.alternate-cluster-kubeconfig}' |\
  base64 -d > "${primary_cluster_kcfg}" || \
  error "Unable to read the secret alternate-cluster-access from uipath-infra namespace"

  kubectl --kubeconfig "${primary_cluster_kcfg}" get secret "rook-ceph-ingress-ca-key" -n "${istio_namespace}" -o jsonpath="{.data.ca_key}" | base64 -d > "${tmp_dir}/ca_key"
  kubectl --kubeconfig "${primary_cluster_kcfg}" get secret "rook-ceph-ingress-ca-key" -n "${istio_namespace}" -o jsonpath="{.data.ca_key_pass}" | base64 -d > "${tmp_dir}/ca_key_pass"
  kubectl --kubeconfig "${primary_cluster_kcfg}" get secret "rook-ceph-ingress-ca" -n "${rook_ceph_namespace}" -o jsonpath="{.data.cabundle}" | base64 -d > "${tmp_dir}/ca_bundle.crt"

  # Create new TLS keypair and sign with CA and pass from primary cluster
  create_self_issued_certificate "objectstore.${ALTERNATIVE_FQDN}" "${ceph_ca_file_path}" \
  "${ceph_tls_file_path}" "${ceph_key_file_path}" "36501" \
  "36500" "36400" 0 "${tmp_dir}" "${tmp_dir}"
  add_empty_line_in_file_if_not_exists "${ceph_tls_file_path}"
  
  merge_ca_and_tls_path="${tmp_dir}/${merged_ca_tls_file}"
  cat "${ceph_tls_file_path}" "${ceph_ca_file_path}" >> "${merge_ca_and_tls_path}"

  # create secret in istio namespace for ingress
  # TODO Investigate ca_cert_name issue
  kubectl create secret generic "rook-ceph-replication-tls" \
    --from-file="${ca_cert_name}"="${ceph_ca_file_path}" \
    --from-file="${tls_cert_name}"="${merge_ca_and_tls_path}" \
    --from-file="${tls_key_name}"="${ceph_key_file_path}" \
    --namespace "${istio_namespace}" \
    --dry-run=client -o yaml | kubectl apply -f - &>>"${dev_null}"

  # cabundle secret in rook-ceph namespace is hanlded by create-secrets.sh job in rook-ceph-object-store helm chart.
}

#MAIN
function main() {
  load_os_utils "${MODULES_DIR}"
  # enable rke2 install/upgrade in package manager (yum,apt etc)
  enable_rke2_package_upgrade

  info "${LANG_INFRA_STARTED_PREFIX} ${LANG_K8S_INSTALL_MESSAGE}"
  # Use trap to append extra error messages to functions imported in this script
  old_trap="$(get_existing_trap "EXIT")"
  #shellcheck disable=SC2064
  trap "error_without_exit \"${LANG_INFRA_FAILED_PREFIX} ${LANG_K8S_INSTALL_MESSAGE}\"; ${old_trap}" EXIT

  # Check the current NODE_ROLE and reject the installation if it's opposite of the existing role
  if [[ ${NODE_ROLE} == "agent" ]];then
    [[ "$(systemctl is-enabled rke2-server)" == "enabled" ]] && error_with_code "11018"
  else
    [[ "$(systemctl is-enabled rke2-agent)" == "enabled" ]] && error_with_code "11019"
  fi

  UPDATE_FQDN=$(get_cluster_config "update_fqdn")
  [[ "${UPDATE_FQDN}" == "true" && ${NODE_ROLE} == "none" ]] && update_cluster_config "update_fqdn_state" "INFRA_STARTED"

  parse_args "$@"
  parse_input_file

  cd "${SCRIPT_DIR}"

  send_checkpoint_telemetry "Infra: Input Parsed"
  send_input_telemetry
  send_hardware_telemetry

  if [[ ${INSTALL_TYPE} == "offline" ]]; then
    DOCKER_REGISTRY_URL="localhost:${DOCKER_REGISTRY_NODEPORT}"
  fi

  populate_node_labels
  generate_rke2_config
  install_rke2
  enable_graceful_node_shutdown

  send_checkpoint_telemetry "Infra: RKE2 Installed"

  if [[ ${NODE_TYPE} == "gpu" ]]; then
    info "${LANG_INFRA_STARTED_PREFIX} ${LANG_GPU_SETUP_MESSAGE}";
    configure_gpu
    info "${LANG_INFRA_COMPLETED_PREFIX} ${LANG_GPU_SETUP_MESSAGE}";
  fi
  if [[ ${NODE_ROLE} == "agent" ]]; then
    test_agent_api

    local hostname_lower
    hostname_lower="$(hostname | awk '{print tolower($0)}')"

    stop_disk_scheduling "${hostname_lower}" "${NODE_ROLE}"
  else
    test_server_api
  fi
  #shellcheck disable=SC2064
  trap "${old_trap}" EXIT
  info "${LANG_INFRA_COMPLETED_PREFIX} ${LANG_K8S_INSTALL_MESSAGE}"

  if [[ ${NODE_ROLE} != "agent" && "${ETCD_RESTORE}" != "true" ]]; then
    create_istio_namespace
    create_uipath_namespace
    create_uipath_infra_namespace
    # Extract kubeconfig for primary cluster in A/A mode, and setup ceph replication TLS.
    if [[ $ACTIVE_ACTIVE == "true" ]]; then
      if [[ ${DO_UPGRADE:-0} != "1" ]]; then
        if [[ $PRIMARY_CLUSTER == "false" ]]; then
          process_primary_cluster_kcfg
          setup_ceph_tls_secondary
        else
          setup_ceph_tls_primary
        fi
      fi
      # Update coredns to be able to resolve the traffic manager fqdn to its own private cluster fqdn
      rke2_coredns_update "true" "${ALTERNATIVE_FQDN}" "${CLUSTER_FQDN}"
    else
      rke2_coredns_update "false"
    fi
  fi

  # with restore flag true only rke2 will be installed
  if [[ "${NODE_ROLE}" == "none" && "${ETCD_RESTORE}" != "true" ]]; then

    store_cluster_configuration

    create_priority_classes

    info "${LANG_INFRA_STARTED_PREFIX} ${LANG_STORAGE_SETUP_MESSAGE}"
    longhorn_prereq
    install_longhorn
    test_longhorn
    if [[ ${INSTALL_TYPE} == "offline" ]]; then
      install_docker_registry
    fi
    info "${LANG_INFRA_COMPLETED_PREFIX} ${LANG_STORAGE_SETUP_MESSAGE}"

    info "${LANG_INFRA_STARTED_PREFIX} ${LANG_ROUTING_SETUP_MESSAGE}"
    install_istio
    configure_istio
    validate_subdomain_resolution
    test_istio_installation
 

    [[ "${INSTALLSTANDALONEAICENTER}" == 0 ]] && store_identity_certificate

    store_additional_ca_certs

    info "${LANG_INFRA_COMPLETED_PREFIX} ${LANG_ROUTING_SETUP_MESSAGE}"
  fi

  send_checkpoint_telemetry "Infra: Longhorn and Registry Components Installed"


  #Fix NFS domain https://longhorn.io/kb/troubleshooting-rwx-mount-ownership-set-to-nobody/
  echo "Domain = localdomain" >> /etc/idmapd.conf || error_with_code "11020"
  nfsidmap -c || error_with_code "11021"

  set_output

  send_checkpoint_telemetry "Infra: Success"

  [[ "${UPDATE_FQDN}" == "true" && ${NODE_ROLE} == "none" ]] && update_cluster_config "update_fqdn_state" "INFRA_DONE"

  if [[ ${NODE_ROLE} != "agent" && "${ETCD_RESTORE}" != "true" ]]; then
    local tls_secret
    tls_secret=$(get_tls_certs)

    echo "${tls_secret}" | jq -r ".data.\"ca.crt\"" | base64 -d | cat > /etc/pki/ca-trust/source/anchors/ca.crt
    update-ca-trust
  fi

  # disable rke2 install/upgrade in package manager (yum,apt etc)
  disable_rke2_package_upgrade

  info "Infra installed successfully"
}

main "$@"
