#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#GLOBALS
export PATH="$PATH:/usr/local/sbin:/sbin:/bin:/usr/sbin:/usr/bin:/root/bin:/usr/local/bin"

SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  echo "Could not determine script path"
  exit 1
fi

MESSAGE_FILE="${SCRIPT_DIR}/../readiness_message_en.json"
if [[ ! -f "${MESSAGE_FILE}" ]] ; then
  echo "Certificate Readiness message file not present"
  exit 1
fi

MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  echo "Could not determine Modules path"
  exit 1
fi

export MODULES_DIR

if [[ -f "${MODULES_DIR}/utils.sh" ]] ; then
  #shellcheck disable=SC1091,SC1090
  source "${MODULES_DIR}/utils.sh"
else
  echo "Common utils not found at ${MODULES_DIR}/utils.sh"
  exit 1
fi

if [[ -f "${MODULES_DIR}/certificate_utils.sh" ]] ; then
  #shellcheck disable=SC1091,SC1090
  source "${MODULES_DIR}/certificate_utils.sh"
else
  echo "Certificate utils not found at ${MODULES_DIR}/certificate_utils.sh"
  exit 1
fi

load_os_utils "${MODULES_DIR}" &>/dev/null

EXIT_CODE=0

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Run certificates checks"
  echo
  echo "Arguments"
  echo "  -h|--help                          Show this help message and exit"
  echo "  -d|--debug                         Increase logging verbosity to show all debug logs"
  echo  
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      echo "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    -*|*) # unsupported flags
      # display_usage
      # echo "Error: Unsupported flag $1"
      # exit 1
      shift
      ;;
    esac
  done
}

function check_if_required_packages_available() {
    if ! check_package_or_command openssl;
    then
        echo "'openssl' not found in PATH or is not installed. Make sure 'openssl' is installed and is available in the PATH"
        exit 1
    fi
}

function validate_certificates() {
    parse_fqdn_values

    parse_certificate_file_paths

    if [[ -s "${CERT_CA_FILE_PATH}" && -s "${CERT_TLS_FILE_PATH}" && -s "${CERT_KEY_FILE_PATH}" ]]; 
    then
        #shellcheck disable=SC2005
        echo "$(get_readiness_message ".certificates.start_info" "${MESSAGE_FILE}" "Server")"
        cert_validation_status=$(validate_server_certificate "${CERT_CA_FILE_PATH}" "${CERT_TLS_FILE_PATH}" "${CERT_KEY_FILE_PATH}" "${CLUSTER_FQDN}" "${ARGO_SUBDOMAIN}" "${OBJECTSTORE_SUBDOMAIN}" "${DOCKER_REGISTRY_SUBDOMAIN}" "${MONITORING_SUBDOMAIN}" "${INSIGHTS_SUBDOMAIN}")
        if [[ "${cert_validation_status}" == "success" ]];
        then
            add_test_result "PASSED" "$(get_readiness_message ".certificates.validation_info" "${MESSAGE_FILE}" "Server")"
        elif [[ "${cert_validation_status}" != "warning" ]]; then
            add_test_result "FAILED" "$(get_readiness_message ".certificates.server_cert_validation_error" "${MESSAGE_FILE}" "${CLUSTER_FQDN}")"
        fi
        echo -e "$(get_readiness_message ".certificates.end_info" "${MESSAGE_FILE}" "Server")\n"
    else
        echo -e "$(get_readiness_message ".certificates.start_info" "${MESSAGE_FILE}" "Server")"
        add_test_result "INFO" "$(get_readiness_message ".certificates.path_missing_info" "${MESSAGE_FILE}" "Server" )" "$(get_readiness_mitigation ".certificates.path_missing_info" "${MESSAGE_FILE}")"
        echo -e "$(get_readiness_message ".certificates.end_info" "${MESSAGE_FILE}" "Server")\n"
    fi

    parse_token_signing_certs_path
    echo -e "$(get_readiness_message ".certificates.start_info" "${MESSAGE_FILE}" "token signing")"
    if [[ -s "${TOKEN_SIGNING_CERT}" && -n "${TOKEN_SIGNING_PASS}" ]]; then
        #shellcheck disable=SC2005
        cert_validation_status=$(validate_signing_certificate "${TOKEN_SIGNING_CERT}" "${TOKEN_SIGNING_PASS}")
        if [[ "${cert_validation_status}" == "success" ]];
        then
            add_test_result "PASSED" "$(get_readiness_message ".certificates.validation_info" "${MESSAGE_FILE}" "Token signing")"
        else
            add_test_result "FAILED"  "$(get_readiness_message ".certificates.validation_error" "${MESSAGE_FILE}" "token signing")"
        fi
    elif [[ -s "${TOKEN_SIGNING_CERT}" && -z "${TOKEN_SIGNING_PASS}" ]]; then
        add_test_result "INFO" "$(get_readiness_message ".certificates.token_signing_password_missing" "${MESSAGE_FILE}")" "$(get_readiness_mitigation ".certificates.token_signing_password_missing" "${MESSAGE_FILE}")"
    else
        add_test_result "INFO" "$(get_readiness_message ".certificates.path_missing_info" "${MESSAGE_FILE}" "Token signing" )" "$(get_readiness_mitigation ".certificates.path_missing_info" "${MESSAGE_FILE}")"
    fi
    echo -e "$(get_readiness_message ".certificates.end_info" "${MESSAGE_FILE}" "token signing")\n"
}

function main() {
  parse_long_args "$@"
  check_if_required_packages_available
  print_center "Certificates Validation"
  validate_certificates
  print_center ""
}

main "$@"
exit "${EXIT_CODE}"