#!/bin/bash

SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)

if [[ -z "${SCRIPT_DIR}" ]]; then
    echo "Script dir not present"
    exit 1
fi

ROOT_DIR="/opt/UiPathAutomationSuite/UiPath_Installer"

HTTP_PROXY=""
HTTPS_PROXY=""
NO_PROXY=""

#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}/utils.sh"

function prepareEnvironment() {
  local installerRootDir="$1"

  if [[ ! -f "${installerRootDir}/Modules/utils.sh" ]]; then
    echo "Installer utils file not present at ${installerRootDir}/Modules/utils.sh"
    exit 1
  fi
  
  source "${installerRootDir}/Modules/utils.sh"

  HTTP_PROXY="http_proxy=$(env | grep "^http_proxy" | awk -F= '{print $2}' | xargs )"
  HTTPS_PROXY="https_proxy=$(env | grep "^https_proxy" | awk -F= '{print $2}' | xargs )"
  NO_PROXY="no_proxy=$(env | grep "^no_proxy" | awk -F= '{print $2}' | xargs )"
  cp /etc/wgetrc /etc/tmp_wgetrc
}

function enabledProxyEnvVariable() {
  export ${HTTP_PROXY}
  export ${HTTPS_PROXY}
  export ${NO_PROXY}
}

function enabledWgetProxy() {
  cp /etc/tmp_wgetrc /etc/wgetrc
}

function disabledProxyEnvVariable() {
  unset http_proxy
  unset https_proxy
  unset no_proxy
}

function disabledWgetProxy() {
  sed '/^http_proxy/s//#&/' /etc/wgetrc >/etc/tmp
  mv /etc/tmp /etc/wgetrc
  sed '/^https_proxy/s//#&/' /etc/wgetrc >/etc/tmp
  mv /etc/tmp /etc/wgetrc
  sed '/^no_proxy/s//#&/' /etc/wgetrc >/etc/tmp
  mv /etc/tmp /etc/wgetrc
}

function whenProxyIsDisabled() {
  local installerRootDir="$1"
  local deployemtType="$2"
  local nodeType="$3"
  local inputFile="$4"
  local exitStatus messagePresent
  exitStatus="true"

  "${installerRootDir}/validateUiPathASReadiness.sh" -i ${inputFile} -n ${nodeType} --install-type ${deployemtType} > /tmp/proxy.log  
  messagePresent="$(grep "Skipping proxy validation, as proxy is disabled" /tmp/proxy.log)"
  if [[ -z ${messagePresent} ]]; then
    exitStatus="false";
  fi

  messagePresent "1648583" "${exitStatus}"
}

function whenProxyConfigurationIsValidAndproxyServerAccessable() {
  local installerRootDir="$1"
  local deployemtType="$2"
  local nodeType="$3"
  local inputFile="$4"

  local exitStatus httpProxyStatus httpsProxyStatus noProxyStatus proxy_rechable

  exitStatus="true"
  "${installerRootDir}/validateUiPathASReadiness.sh" -i ${inputFile} -n ${nodeType} --install-type ${deployemtType} > /tmp/proxy.log
  httpProxyStatus="$(grep "http_proxy is set properly" /tmp/proxy.log)"
  httpsProxyStatus="$(grep "https_proxy is set properly" /tmp/proxy.log)"
  noProxyStatus="$(grep "no_proxy is set properly" /tmp/proxy.log)"
  proxy_rechable="$(grep "Able to resolve proxy server from this machine" /tmp/proxy.log)"
  
  if [[ -z "${httpProxyStatus}" || -z "${httpsProxyStatus}" \
    || -z "${noProxyStatus}" || -z "${proxy_rechable}" ]]; then
    exitStatus="false"
  fi

  appendOutput "1648584" "${exitStatus}"
}

function whenProxyConfigurationIsValidAndproxyServerNotAccessable() {
  local installerRootDir="$1"
  local deployemtType="$2"
  local nodeType="$3"
  local inputFile="$4"
  local exitStatus httpProxyStatus httpsProxyStatus noProxyStatus proxy_rechable

  exitStatus="false"
  "${installerRootDir}/validateUiPathASReadiness.sh" -i ${inputFile} -n ${nodeType} --install-type ${deployemtType} > /tmp/proxy.log
  if [[ $? -eq 1 ]]; then
    exitStatus="true"
    httpProxyStatus="$(grep "http_proxy is set properly" /tmp/proxy.log)"
    httpsProxyStatus="$(grep "https_proxy is set properly" /tmp/proxy.log)"
    noProxyStatus="$(grep "no_proxy is set properly" /tmp/proxy.log)"
    proxy_rechable="$(grep "Unable to resolve proxy server from this machine" /tmp/proxy.log)"
  fi
  
  if [[ -z "${httpProxyStatus}" || -z "${httpsProxyStatus}" \
    || -z "${noProxyStatus}" || -z "${proxy_rechable}" ]]; then
    exitStatus="false"
  fi

  appendOutput "1648585" "${exitStatus}"
}

function whenProxyConfigurationIsInvalid() {
  local installerRootDir="$1"
  local deployemtType="$2"
  local nodeType="$3"
  local inputFile="$4"
  local curlExitStatus wgetExitStatus httpProxyStatus httpsProxyStatus noProxyStatus
  
  curlExitStatus="true"
  wgetExitStatus="true"
  disabledProxyEnvVariable
  "${installerRootDir}/validateUiPathASReadiness.sh" -i ${inputFile} -n ${nodeType} --install-type ${deployemtType} > /tmp/proxy.log
  if [[ $? -eq 1 ]]; then
    httpProxyStatus="$(grep "http_proxy is not specified" /tmp/proxy.log)"
    httpsProxyStatus="$(grep "https_proxy is not specified" /tmp/proxy.log)"
    noProxyStatus="$(grep "no_proxy is not specified" /tmp/proxy.log)"
  fi
  enabledProxyEnvVariable
  if [[ -z "${httpProxyStatus}" || -z "${httpsProxyStatus}" || -z "${noProxyStatus}" ]]; then
    curlExitStatus="false"
  fi

  httpProxyStatus=""
  httpsProxyStatus=""
  noProxyStatus=""
  disabledWgetProxy
  "${installerRootDir}/validateUiPathASReadiness.sh" -i ${inputFile} -n ${nodeType} --install-type ${deployemtType} > /tmp/proxy.log
  if [[ $? -eq 1 ]]; then
    httpProxyStatus="$(grep "http_proxy is not specified" /tmp/proxy.log)"
    httpsProxyStatus="$(grep "https_proxy is not specified" /tmp/proxy.log)"
    noProxyStatus="$(grep "no_proxy is not specified" /tmp/proxy.log)"
  fi
  enabledWgetProxy
  if [[ -z "${httpProxyStatus}" || -z "${httpsProxyStatus}" || -z "${noProxyStatus}" ]]; then
    wgetExitStatus="false"
  fi

  if [[ "${curlExitStatus}" == "false" || "${wgetExitStatus}" == "false" ]]; then
    appendOutput "1648586" "false"
  else
    appendOutput "1648586" "true"
  fi
}

function whenNoProxyConfigurationMissingClusterFQDN() {
  local installerRootDir="$1"
  local deployemtType="$2"
  local nodeType="$3"
  local inputFile="$4"
  local exitStatus httpProxyStatus httpsProxyStatus noProxyStatus fqdnStatus noProxy fqdn

  fqdn=$(jq ".fqdn" ${inputFile} | xargs)
  noProxy=$(echo ${NO_PROXY} | sed s/"$fqdn"// | sed s/"alm.$fqdn"//)
  sed '/^no_proxy/s//#&/' /etc/wgetrc >/etc/tmp
  mv /etc/tmp /etc/wgetrc

  export ${noProxy}
  echo ${noProxy} >> /etc/wgetrc

  exitStatus="true"
  "${installerRootDir}/validateUiPathASReadiness.sh" -i ${inputFile} -n ${nodeType} --install-type ${deployemtType} > /tmp/proxy.log
  if [[ $? -eq 1 ]]; then
    httpProxyStatus="$(grep "http_proxy is set properly" /tmp/proxy.log)"
    httpsProxyStatus="$(grep "https_proxy is set properly" /tmp/proxy.log)"
    noProxyStatus="$(grep "no_proxy is set properly" /tmp/proxy.log)"
    fqdnStatus="$(grep "fqdn configuration is missing from no_proxy" /tmp/proxy.log)"
  fi

  if [[ -z "${httpProxyStatus}" || -z "${httpsProxyStatus}" \
    || -z "${noProxyStatus}" || -z "${fqdnStatus}" ]]; then
    exitStatus="false"
  fi

  appendOutput "1648587" "${exitStatus}"

  enabledProxyEnvVariable
  enabledWgetProxy
}

function executeTest() {
  local proxyFlag
  prepareEnvironment "${ROOT_DIR}"
  proxyFlag=$(jq ".proxy.enabled" ${ROOT_DIR}/input.json | xargs)
  if [[ "${VALIDATE_STAGE}" == "proxyServerRechable" && "${proxyFlag}" == "true" ]]; then
    whenProxyConfigurationIsValidAndproxyServerAccessable "${ROOT_DIR}" "${DEPLOYMENT_TYPE}" "${NODE_TYPE}" "${ROOT_DIR}/input.json"
    whenProxyConfigurationIsInvalid "${ROOT_DIR}" "${DEPLOYMENT_TYPE}" "${NODE_TYPE}" "${ROOT_DIR}/input.json"
    whenNoProxyConfigurationMissingClusterFQDN "${ROOT_DIR}" "${DEPLOYMENT_TYPE}" "${NODE_TYPE}" "${ROOT_DIR}/input.json"
  elif [[ "${VALIDATE_STAGE}" == "proxyServerUnRechable"  && "${proxyFlag}" == "true" ]]; then
    whenProxyConfigurationIsValidAndproxyServerNotAccessable "${ROOT_DIR}" "${DEPLOYMENT_TYPE}" "${NODE_TYPE}" "${ROOT_DIR}/input.json"
  elif [[ -z "${proxyFlag}" || "${proxyFlag}" == "false" ]]; then
    whenProxyIsDisabled "${ROOT_DIR}" "${DEPLOYMENT_TYPE}" "${NODE_TYPE}" "${ROOT_DIR}/input.json"
  fi
}

parse_args "$@"
executeTest
